from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of Apache Druid.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port that Apache Druid is running on.',
        'required': True,
        'label': 'Port'
    },
    path={
        'type': ARG_TYPE.STR,
        'description': 'The query path.',
        'required': True,
        'label': 'path'
    },
    scheme={
        'type': ARG_TYPE.STR,
        'description': 'The URI schema. This parameter is optional and the default will be http.',
        'required': False,
        'label': 'Scheme'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with Apache Druid. This parameter is optional.',
        'required': False,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password used to authenticate with Apache Druid. This parameter is optional.',
        'required': False,
        'label': 'password',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    host='localhost',
    port=8888,
    path='/druid/v2/sql/',
    scheme='http'
)



from mindsdb_sql_parser.ast import (
    Identifier,
    Constant,
    Select,
    Star,
)
from mindsdb.api.executor.planner.steps import (
    GetPredictorColumns,
    GetTableColumns,
)

from mindsdb.api.executor.sql_query.result_set import ResultSet, Column

from .base import BaseStepCall


class GetPredictorColumnsCall(BaseStepCall):

    bind = GetPredictorColumns

    def call(self, step):

        mindsdb_database_name = 'mindsdb'

        predictor_name = step.predictor.parts[-1]
        dn = self.session.datahub.get(mindsdb_database_name)
        columns = [col['name'] for col in dn.get_table_columns(predictor_name)]

        data = ResultSet()
        for column_name in columns:
            data.add_column(Column(
                name=column_name,
                table_name=predictor_name,
                database=mindsdb_database_name
            ))
        return data


class GetTableColumnsCall(BaseStepCall):

    bind = GetTableColumns

    def call(self, step):

        table = step.table
        dn = self.session.datahub.get(step.namespace)
        ds_query = Select(from_table=Identifier(table), targets=[Star()], limit=Constant(0))

        data, columns_info = dn.query(ds_query, session=self.session)

        data = ResultSet()
        for column in columns_info:
            data.add_column(Column(
                name=column['name'],
                type=column.get('type'),
                table_name=table,
                database=self.context.get('database')
            ))
        return data



from typing import Dict, List, Text

from flask import request
import msal

from mindsdb.integrations.utilities.handlers.auth_utilities.exceptions import AuthException
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSGraphAPIDelegatedPermissionsManager:
    """
    The class for managing the delegated permissions for the Microsoft Graph API.
    """
    def __init__(
        self,
        client_id: Text,
        client_secret: Text,
        tenant_id: Text,
        cache: msal.SerializableTokenCache,
        scopes: List = ["https://graph.microsoft.com/.default"],
        code: Text = None,
    ) -> None:
        """
        Initializes the delegated permissions manager.

        Args:
            client_id (Text): The client ID of the application registered in Microsoft Entra ID.
            client_secret (Text): The client secret of the application registered in Microsoft Entra ID.
            tenant_id (Text): The tenant ID of the application registered in Microsoft Entra ID.
            cache (msal.SerializableTokenCache): The token cache for storing the access token.
            scopes (List): The scopes for the Microsoft Graph API.
            code (Text): The authentication code for acquiring the access token.
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.cache = cache
        self.scopes = scopes
        self.code = code

        # Set the redirect URI based on the request origin.
        # If the request origin is 127.0.0.1 (localhost), replace it with localhost.
        # This is done because the only HTTP origin allowed in Microsoft Entra ID app registration is localhost.
        try:
            request_origin = request.headers.get('ORIGIN') or (request.scheme + '://' + request.host)
            if not request_origin:
                raise AuthException('Request origin could not be determined!')
        except RuntimeError:
            # if it is outside of request context (streaming in agent)
            request_origin = ''

        request_origin = request_origin.replace('127.0.0.1', 'localhost') if 'http://127.0.0.1' in request_origin else request_origin
        self.redirect_uri = request_origin + '/verify-auth'

    def get_access_token(self) -> Text:
        """
        Retrieves an access token for the Microsoft Graph API.
        If a valid access token is found in the cache, it is returned.
        Otherwise, the authentication flow is executed.

        Returns:
            Text: The access token for the Microsoft Graph API.
        """
        # Check if a valid access token is already in the cache for the signed-in user.
        msal_app = self._get_msal_app()
        accounts = msal_app.get_accounts()

        if accounts:
            response = msal_app.acquire_token_silent(self.scopes, account=accounts[0])
            if "access_token" in response:
                return response['access_token']

        # If no valid access token is found in the cache, run the authentication flow.
        response = self._execute_ms_graph_api_auth_flow()

        if "access_token" in response:
            return response['access_token']
        # If no access token is returned, raise an exception.
        # This is the expected behaviour when the user attempts to authenticate for the first time.
        else:
            raise AuthException(
                f'Error getting access token: {response.get("error_description")}',
                auth_url=response.get('auth_url')
            )

    def _get_msal_app(self) -> msal.ConfidentialClientApplication:
        """
        Returns an instance of the MSAL ConfidentialClientApplication.

        Returns:
            msal.ConfidentialClientApplication: An instance of the MSAL ConfidentialClientApplication.
        """
        return msal.ConfidentialClientApplication(
            self.client_id,
            authority=f"https://login.microsoftonline.com/{self.tenant_id}",
            client_credential=self.client_secret,
            token_cache=self.cache,
        )

    def _execute_ms_graph_api_auth_flow(self) -> Dict:
        """
        Executes the authentication flow for the Microsoft Graph API.
        If the authentication code is provided, the token is acquired by authorization code.
        Otherwise, the authorization request URL is returned.

        Raises:
            AuthException: If the authentication code is not provided

        Returns:
            Dict: The response from the Microsoft Graph API authentication flow.
        """
        msal_app = self._get_msal_app()

        # If the authentication code is provided, acquire the token by authorization code.
        if self.code:
            response = msal_app.acquire_token_by_authorization_code(
                code=self.code,
                scopes=self.scopes,
                redirect_uri=self.redirect_uri
            )

            return response

        # If the authentication code is not provided, get the authorization request URL.
        else:
            auth_url = msal_app.get_authorization_request_url(
                scopes=self.scopes,
                redirect_uri=self.redirect_uri
            )

            raise AuthException(f'Authorisation required. Please follow the url: {auth_url}', auth_url=auth_url)



import os

from langchain.agents import AgentType
from langchain_openai import OpenAIEmbeddings

from types import MappingProxyType
from mindsdb.integrations.handlers.openai_handler.constants import (
    CHAT_MODELS as OPEN_AI_CHAT_MODELS,
)

SUPPORTED_PROVIDERS = {
    "openai",
    "anthropic",
    "anyscale",
    "litellm",
    "ollama",
    "nvidia_nim",
    "vllm"
}
# Chat models
ANTHROPIC_CHAT_MODELS = (
    "claude-3-opus-20240229",
    "claude-3-sonnet-20240229",
    "claude-3-haiku-20240307",
    "claude-2.1",
    "claude-2.0",
    "claude-instant-1.2",
)

OLLAMA_CHAT_MODELS = (
    "gemma",
    "llama2",
    "mistral",
    "mixtral",
    "llava",
    "neural-chat",
    "codellama",
    "dolphin-mixtral",
    "qwen",
    "llama2-uncensored",
    "mistral-openorca",
    "deepseek-coder",
    "nous-hermes2",
    "phi",
    "orca-mini",
    "dolphin-mistral",
    "wizard-vicuna-uncensored",
    "vicuna",
    "tinydolphin",
    "llama2-chinese",
    "openhermes",
    "zephyr",
    "nomic-embed-text",
    "tinyllama",
    "openchat",
    "wizardcoder",
    "phind-codellama",
    "starcoder",
    "yi",
    "orca2",
    "falcon",
    "starcoder2",
    "wizard-math",
    "dolphin-phi",
    "nous-hermes",
    "starling-lm",
    "stable-code",
    "medllama2",
    "bakllava",
    "codeup",
    "wizardlm-uncensored",
    "solar",
    "everythinglm",
    "sqlcoder",
    "nous-hermes2-mixtral",
    "stable-beluga",
    "yarn-mistral",
    "samantha-mistral",
    "stablelm2",
    "meditron",
    "stablelm-zephyr",
    "magicoder",
    "yarn-llama2",
    "wizard-vicuna",
    "llama-pro",
    "deepseek-llm",
    "codebooga",
    "mistrallite",
    "dolphincoder",
    "nexusraven",
    "open-orca-platypus2",
    "all-minilm",
    "goliath",
    "notux",
    "alfred",
    "megadolphin",
    "xwinlm",
    "wizardlm",
    "duckdb-nsql",
    "notus",
)

NVIDIA_NIM_CHAT_MODELS = (
    "microsoft/phi-3-mini-4k-instruct",
    "mistralai/mistral-7b-instruct-v0.2",
    "writer/palmyra-med-70b",
    "mistralai/mistral-large",
    "mistralai/codestral-22b-instruct-v0.1",
    "nvidia/llama3-chatqa-1.5-70b",
    "upstage/solar-10.7b-instruct",
    "google/gemma-2-9b-it",
    "adept/fuyu-8b",
    "google/gemma-2b",
    "databricks/dbrx-instruct",
    "meta/llama-3_1-8b-instruct",
    "microsoft/phi-3-medium-128k-instruct",
    "01-ai/yi-large",
    "nvidia/neva-22b",
    "meta/llama-3_1-70b-instruct",
    "google/codegemma-7b",
    "google/recurrentgemma-2b",
    "google/gemma-2-27b-it",
    "deepseek-ai/deepseek-coder-6.7b-instruct",
    "mediatek/breeze-7b-instruct",
    "microsoft/kosmos-2",
    "microsoft/phi-3-mini-128k-instruct",
    "nvidia/llama3-chatqa-1.5-8b",
    "writer/palmyra-med-70b-32k",
    "google/deplot",
    "meta/llama-3_1-405b-instruct",
    "aisingapore/sea-lion-7b-instruct",
    "liuhaotian/llava-v1.6-mistral-7b",
    "microsoft/phi-3-small-8k-instruct",
    "meta/codellama-70b",
    "liuhaotian/llava-v1.6-34b",
    "nv-mistralai/mistral-nemo-12b-instruct",
    "microsoft/phi-3-medium-4k-instruct",
    "seallms/seallm-7b-v2.5",
    "mistralai/mixtral-8x7b-instruct-v0.1",
    "mistralai/mistral-7b-instruct-v0.3",
    "google/paligemma",
    "google/gemma-7b",
    "mistralai/mixtral-8x22b-instruct-v0.1",
    "google/codegemma-1.1-7b",
    "nvidia/nemotron-4-340b-instruct",
    "meta/llama3-70b-instruct",
    "microsoft/phi-3-small-128k-instruct",
    "ibm/granite-8b-code-instruct",
    "meta/llama3-8b-instruct",
    "snowflake/arctic",
    "microsoft/phi-3-vision-128k-instruct",
    "meta/llama2-70b",
    "ibm/granite-34b-code-instruct",
)

# Define a read-only dictionary mapping providers to their models
PROVIDER_TO_MODELS = MappingProxyType(
    {
        "anthropic": ANTHROPIC_CHAT_MODELS,
        "ollama": OLLAMA_CHAT_MODELS,
        "openai": OPEN_AI_CHAT_MODELS,
        "nvidia_nim": NVIDIA_NIM_CHAT_MODELS,
    }
)

ASSISTANT_COLUMN = "answer"
CONTEXT_COLUMN = "context"
TRACE_ID_COLUMN = "trace_id"
DEFAULT_AGENT_TIMEOUT_SECONDS = 300
# These should require no additional arguments.
DEFAULT_AGENT_TOOLS = []
DEFAULT_AGENT_TYPE = AgentType.CONVERSATIONAL_REACT_DESCRIPTION
DEFAULT_MAX_ITERATIONS = 10
DEFAULT_MAX_TOKENS = 8096
DEFAULT_MODEL_NAME = "gpt-4o"
DEFAULT_TEMPERATURE = 0.0
USER_COLUMN = "question"
DEFAULT_EMBEDDINGS_MODEL_PROVIDER = "openai"
DEFAULT_EMBEDDINGS_MODEL_CLASS = OpenAIEmbeddings
DEFAULT_TIKTOKEN_MODEL_NAME = os.getenv('DEFAULT_TIKTOKEN_MODEL_NAME', 'gpt-4')
AGENT_CHUNK_POLLING_INTERVAL_SECONDS = os.getenv('AGENT_CHUNK_POLLING_INTERVAL_SECONDS', 1.0)



from typing import Optional, Dict
import dill
import dspy


from dspy import ColBERTv2
from dspy.teleprompt import BootstrapFewShot

from mindsdb.interfaces.llm.llm_controller import LLMDataController
import pandas as pd
from mindsdb.interfaces.agents.constants import (
    DEFAULT_MODEL_NAME
)
from mindsdb.interfaces.agents.langchain_agent import (
    get_llm_provider, get_embedding_model_provider
)
from mindsdb.integrations.utilities.rag.settings import DEFAULT_RAG_PROMPT_TEMPLATE
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.interfaces.storage.model_fs import HandlerStorage, ModelStorage
from mindsdb.utilities import log


_PARSING_ERROR_PREFIX = 'An output parsing error occured'

logger = log.getLogger(__name__)

START_URL = 'http://20.102.90.50:2017/wiki17_abstracts'
DEMOS = 4
DF_EXAMPLES = 25
PRIME_MULTIPLIER = 31


class DSPyHandler(BaseMLEngine):

    """
    This is a MindsDB integration for the DSPy library, which provides a unified interface for interacting with
    various large language models (LLMs) and self improving prompts for these LLMs.

    Supported LLM providers:
        - OpenAI

    Supported standard tools:
        - python_repl
        - serper.dev search
    """
    name = 'dspy'

    def __init__(
            self,
            model_storage: ModelStorage,
            engine_storage: HandlerStorage,
            **kwargs):
        super().__init__(model_storage, engine_storage, **kwargs)
        # if True, the target column name does not have to be specified at creation time.
        self.generative = True
        self.llm_data_controller = LLMDataController()
        self.model_id = 0

    def calculate_model_id(self, model_name):
        '''
        Based on the model name calculate a unique number to be used as model_id
        '''
        sum_chars = sum((i + 1) * ord(char) * PRIME_MULTIPLIER for i, char in enumerate(model_name))
        return sum_chars

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Dict = None, **kwargs):
        """
        Create a model by initializing the parameters and setting up a DSPy chain with cold start data

        Args:
            target (str): Type of engine
            df (DataFrame): cold start df for DSPy
            args (Dict): Parameters for the model

        Returns:
            None
        """

        args = args['using']
        args['target'] = target
        args['model_name'] = args.get('model_name', DEFAULT_MODEL_NAME)
        args['provider'] = args.get('provider', get_llm_provider(args))
        args['embedding_model_provider'] = args.get('embedding_model', get_embedding_model_provider(args))
        if args.get('mode') == 'retrieval':
            # use default prompt template for retrieval i.e. RAG if not provided
            if "prompt_template" not in args:
                args["prompt_template"] = DEFAULT_RAG_PROMPT_TEMPLATE

        self.model_storage.json_set('args', args)
        if not isinstance(df, type(None)):
            self.model_storage.file_set('cold_start_df', dill.dumps(df.to_dict()))
        else:
            # no cold start df if data not provided
            self.model_storage.file_set('cold_start_df', dill.dumps({}))
        # TODO: temporal workaround: serialize df and args, instead. And recreate chain (with training) every inference call.
        # ideally, we serialize the chain itself to avoid duplicate training.

    def predict(self, df: pd.DataFrame, args: Dict = None) -> pd.DataFrame:
        """
        Predicts a response using DSPy

        Args:
            df (DataFrame): input for the model
            args (Dict): Parameters for the model

        Returns:
            df (DataFrame): response from the model
        """
        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        if 'prompt_template' not in args and 'prompt_template' not in pred_args:
            raise ValueError("This model expects a `prompt_template`, please provide one.")
        # Back compatibility for old models
        args['provider'] = args.get('provider', get_llm_provider(args))
        args['embedding_model_provider'] = args.get('embedding_model', get_embedding_model_provider(args))

        # retrieves llm and pass it around as context
        model = args.get('model_name')
        self.model_id = self.calculate_model_id(model)
        api_key = get_api_key('openai', args, self.engine_storage, strict=False)
        llm = dspy.OpenAI(model=model, api_key=api_key)

        df = df.reset_index(drop=True)

        cold_start_df = pd.DataFrame(dill.loads(self.model_storage.file_get("cold_start_df")))  # fixed in "training"  # noqa
        # gets larger as agent is used more
        self_improvement_df = pd.DataFrame(self.llm_data_controller.list_all_llm_data(self.model_id))
        if self_improvement_df.empty:
            self_improvement_df = pd.DataFrame([{'input': 'dummy_input', 'output': 'dummy_output'}])

        self_improvement_df = self_improvement_df.rename(columns={
            'output': args['target'],
            'input': args['user_column']
        })
        self_improvement_df = self_improvement_df.tail(DF_EXAMPLES)

        # add cold start DF
        self_improvement_df = pd.concat([cold_start_df, self_improvement_df]).reset_index(drop=True)
        chain = self.setup_dspy(self_improvement_df, args)
        output = self.predict_dspy(df, args, chain, llm)  # this stores new traces for self-improvement
        return output

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        tables = ['info']
        return pd.DataFrame(tables, columns=['tables'])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        raise NotImplementedError('Fine-tuning is not supported for LangChain models')

    def setup_dspy(self, df, args):
        """
        Use the default DSPy parameters to set up the chain

        Args:
            df (DataFrame): input to the model
            args (Dict): Parameters for the model

        Returns:
            DSPy chain
        """
        # This is the default language model and retrieval model in DSPy
        colbertv2 = ColBERTv2(url=START_URL)
        dspy.configure(rm=colbertv2)
        dspy_chain = self.create_dspy_chain(df, args)
        return dspy_chain

    def create_dspy_chain(self, df, args):
        """
        Iniialize chain with the llm, add the cold start examples and bootstrap some examples

        Args:
            df (DataFrame): input to the model
            args (Dict): Parameters for the model

        Returns:
            Optimized DSPy chain
        """

        # Initialize the LLM with the API key
        model = args.get('model_name')
        api_key = get_api_key(model, args, self.engine_storage, strict=False)
        llm = dspy.OpenAI(model=model, api_key=api_key)

        # Convert to DSPy Module
        with dspy.context(lm=llm):
            dspy_module = dspy.ReAct(f'{args["user_column"]} -> {args["target"]}')

        # create a list of DSPy examples
        dspy_examples = []

        # TODO: maybe random choose a fixed set of rows
        for i, row in df.iterrows():
            example = dspy.Example(
                question=row[args["user_column"]],
                answer=row[args["target"]]
            ).with_inputs(args["user_column"])
            dspy_examples.append(example)

        # TODO: add the optimizer, maybe the metric
        config = dict(max_bootstrapped_demos=DEMOS, max_labeled_demos=DEMOS)
        metric = dspy.evaluate.metrics.answer_exact_match  # TODO: passage match requires context from prediction... we'll probably modify the signature of ReAct
        teleprompter = BootstrapFewShot(metric=metric, **config)  # TODO: maybe it's better to have this persisted so that the internal state does a better job at optimizing RAG
        with dspy.context(lm=llm):
            optimized = teleprompter.compile(dspy_module, trainset=dspy_examples)  # TODO: check columns have the right name
        return optimized

    def generate_dspy_response(self, question, chain, llm):
        """
        Generate response using DSPy

        Args:
            question (str): question asked as input to the model
            chain: DSPy chain created
            llm: OpenAI model

        Returns:
            Answer to the prompt
        """
        input_dict = {"question": question}
        with dspy.context(lm=llm):
            response = chain(question=input_dict['question'])
        return response.answer

    def predict_dspy(self, df: pd.DataFrame, args: Dict, chain, llm) -> pd.DataFrame:
        """
        Generate response using DSPy

        Args:
            df (DataFrame): contains the input to the model
            args (Dict): parameters of the model
            chain: DSPy chain created
            llm: OpenAI model

        Returns:
            df (DataFrame): Dataframe of responses
        """
        responses = []
        for index, row in df.iterrows():
            question = row[args['user_column']]
            answer = self.generate_dspy_response(question, chain, llm)
            responses.append({'answer': answer, 'question': question})
            # TODO: check this only adds new incoming rows
            self.llm_data_controller.add_llm_data(question, answer, self.model_id)  # stores new traces for use in new calls

        # Set up the evaluator, which can be used multiple times.
        # TODO: use this in the EVALUATE command
        # evaluate = Evaluate(devset=gsm8k_devset, metric=gsm8k_metric, num_threads=4, display_progress=True, display_table=0)

        return pd.DataFrame(responses)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the Couchbase server.",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the Couchbase server.",
        'secret': True
    },
    bucket={
        "type": ARG_TYPE.STR,
        "description": "The database/bucket name to use when connecting with the Couchbase server.",
    },
    connection_string={
        "type": ARG_TYPE.STR,
        "description": "the Connection string to specify the cluster endpoint.",
    },
    scope={
        "type": ARG_TYPE.STR,
        "description": 'The scope use in the query context in Couchbase server. If blank, scope will be "_default".',
    },
)
connection_args_example = OrderedDict(
    connection_string="couchbase://localhost", user="root", password="password", bucket="bucket", scope="example_scope"
)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .slack_handler import (
        SlackHandler as Handler
    )
    import_error = None
except Exception as e:
    Handler = None
    import_error = e 

title = 'Slack'
name = 'slack'
type = HANDLER_TYPE.DATA
icon_path = 'icon.svg'

__all__ = [
    'Handler',
    'version',
    'name',
    'type',
    'title', 
    'description',
    'import_error',
    'icon_path'
]


from http import HTTPStatus

from flask import request
from flask_restx import Resource
from langchain_text_splitters import MarkdownHeaderTextSplitter

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.exceptions import ExecutorException
from mindsdb.api.http.utils import http_error

from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.integrations.utilities.rag.splitters.file_splitter import FileSplitter, FileSplitterConfig
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.knowledge_base.preprocessing.constants import (
    DEFAULT_CONTEXT_DOCUMENT_LIMIT,
    DEFAULT_CRAWL_DEPTH, DEFAULT_WEB_FILTERS,
    DEFAULT_MARKDOWN_HEADERS, DEFAULT_WEB_CRAWL_LIMIT
)
from mindsdb.interfaces.knowledge_base.preprocessing.document_loader import DocumentLoader
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseTable
from mindsdb.utilities import log
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.integrations.utilities.rag.settings import DEFAULT_LLM_MODEL, DEFAULT_RAG_PROMPT_TEMPLATE


from mindsdb_sql_parser.ast import Identifier

logger = log.getLogger(__name__)


@ns_conf.route('/<project_name>/knowledge_bases')
class KnowledgeBasesResource(Resource):
    @ns_conf.doc('list_knowledge_bases')
    @api_endpoint_metrics('GET', '/knowledge_bases')
    def get(self, project_name):
        '''List all knowledge bases'''
        session = SessionController()
        project_controller = ProjectController()
        try:
            _ = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        # KB Controller already returns dict.
        return session.kb_controller.list(project_name)

    @ns_conf.doc('create_knowledge_base')
    @api_endpoint_metrics('POST', '/knowledge_bases')
    def post(self, project_name):
        '''Create a knowledge base'''

        # Check for required parameters.
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in POST body'
            )

        knowledge_base = request.json['knowledge_base']
        # Explicitly require embedding model & vector database.
        required_fields = ['name', 'model']
        for field in required_fields:
            if field not in knowledge_base:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing parameter',
                    f'Must provide "{field}" field in "knowledge_base"'
                )
        if 'storage' in knowledge_base:
            if 'database' not in knowledge_base['storage'] or 'table' not in knowledge_base['storage']:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing parameter',
                    'Must provide "database" and "table" field in "storage" param'
                )

        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        kb_name = knowledge_base.get('name')
        existing_kb = session.kb_controller.get(kb_name, project.id)
        if existing_kb is not None:
            # Knowledge Base must not exist.
            return http_error(
                HTTPStatus.CONFLICT,
                'Knowledge Base already exists',
                f'Knowledge Base with name {kb_name} already exists'
            )

        embedding_model_identifier = None
        if knowledge_base.get('model'):
            embedding_model_identifier = Identifier(parts=[knowledge_base['model']])

        storage = knowledge_base.get('storage')
        embedding_table_identifier = None
        if storage is not None:
            embedding_table_identifier = Identifier(parts=[storage['database'], storage['table']])

        try:
            new_kb = session.kb_controller.add(
                kb_name,
                project.name,
                embedding_model_identifier,
                embedding_table_identifier,
                params=knowledge_base.get('params', {}),
                preprocessing_config=knowledge_base.get('preprocessing')
            )
        except ValueError as e:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Invalid preprocessing configuration',
                str(e)
            )

        return new_kb.as_dict(), HTTPStatus.CREATED


@ns_conf.route('/<project_name>/knowledge_bases/<knowledge_base_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('knowledge_base_name', 'Name of the knowledge_base')
class KnowledgeBaseResource(Resource):
    @ns_conf.doc('get_knowledge_base')
    @api_endpoint_metrics('GET', '/knowledge_bases/knowledge_base')
    def get(self, project_name, knowledge_base_name):
        '''Gets a knowledge base by name'''
        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        existing_kb = session.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )
        return existing_kb.as_dict()

    @ns_conf.doc('update_knowledge_base')
    @api_endpoint_metrics('PUT', '/knowledge_bases/knowledge_base')
    def put(self, project_name: str, knowledge_base_name: str):
        '''Updates a knowledge base with optional preprocessing.'''

        # Check for required parameters
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in PUT body'
            )

        session = SessionController()
        project_controller = ProjectController()

        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        try:
            kb_data = request.json['knowledge_base']

            # Retrieve the knowledge base table for updates
            table = session.kb_controller.get_table(knowledge_base_name, project.id, params=kb_data.get('params'))
            if table is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Knowledge Base not found',
                    f'Knowledge Base with name {knowledge_base_name} does not exist'
                )

            # Set up dependencies for DocumentLoader
            file_controller = FileController()
            file_splitter_config = FileSplitterConfig()
            file_splitter = FileSplitter(file_splitter_config)
            markdown_splitter = MarkdownHeaderTextSplitter(headers_to_split_on=DEFAULT_MARKDOWN_HEADERS)
            mysql_proxy = FakeMysqlProxy()

            # Initialize DocumentLoader with required components
            document_loader = DocumentLoader(
                file_controller=file_controller,
                file_splitter=file_splitter,
                markdown_splitter=markdown_splitter,
                mysql_proxy=mysql_proxy
            )

            # Configure table with dependencies
            table.document_loader = document_loader

            # Update preprocessing configuration if provided
            if 'preprocessing' in kb_data:
                table.configure_preprocessing(kb_data['preprocessing'])

            # Process raw data rows if provided
            if kb_data.get('rows'):
                table.insert_rows(kb_data['rows'])

            # Process files if specified
            if kb_data.get('files'):
                table.insert_files(kb_data['files'])

            # Process web pages if URLs provided
            if kb_data.get('urls'):
                table.insert_web_pages(
                    urls=kb_data['urls'],
                    limit=kb_data.get('limit', DEFAULT_WEB_CRAWL_LIMIT),
                    crawl_depth=kb_data.get('crawl_depth', DEFAULT_CRAWL_DEPTH),
                    filters=kb_data.get('filters', DEFAULT_WEB_FILTERS)
                )

            # Process query if provided
            if kb_data.get('query'):
                table.insert_query_result(kb_data['query'], project_name)

        except ExecutorException as e:
            logger.error(f'Error during preprocessing and insertion: {str(e)}')
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Invalid SELECT query',
                f'Executing "query" failed. Needs to be a valid SELECT statement that returns data: {str(e)}'
            )

        except Exception as e:
            logger.error(f'Error during preprocessing and insertion: {str(e)}')
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Preprocessing Error',
                f'Error during preprocessing and insertion: {str(e)}'
            )

        return '', HTTPStatus.OK

    @ns_conf.doc('delete_knowledge_base')
    @api_endpoint_metrics('DELETE', '/knowledge_bases/knowledge_base')
    def delete(self, project_name: str, knowledge_base_name: str):
        '''Deletes a knowledge base.'''
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        session_controller = SessionController()
        existing_kb = session_controller.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            # Knowledge Base must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

        session_controller.kb_controller.delete(knowledge_base_name, project_name)
        return '', HTTPStatus.NO_CONTENT


def _handle_chat_completion(knowledge_base_table: KnowledgeBaseTable, request):
    # Check for required parameters
    query = request.json.get('query')

    llm_model = request.json.get('llm_model')
    if llm_model is None:
        logger.warning(f'Missing parameter "llm_model" in POST body, using default llm_model {DEFAULT_LLM_MODEL}')

    prompt_template = request.json.get('prompt_template')
    if prompt_template is None:
        logger.warning(f'Missing parameter "prompt_template" in POST body, using default prompt template {DEFAULT_RAG_PROMPT_TEMPLATE}')

    # Get retrieval config, if set
    retrieval_config = request.json.get('retrieval_config', {})
    if not retrieval_config:
        logger.warning('No retrieval config provided, using default retrieval config')

    # add llm model to retrieval config
    if llm_model is not None:
        retrieval_config['llm_model_name'] = llm_model

    # add prompt template to retrieval config
    if prompt_template is not None:
        retrieval_config['rag_prompt_template'] = prompt_template

    # add llm provider to retrieval config if set
    llm_provider = request.json.get('model_provider')
    if llm_provider is not None:
        retrieval_config['llm_provider'] = llm_provider

    # build rag pipeline
    rag_pipeline = knowledge_base_table.build_rag_pipeline(retrieval_config)

    # get response from rag pipeline
    rag_response = rag_pipeline(query)
    response = {
        'message': {
            'content': rag_response.get('answer'),
            'context': rag_response.get('context'),
            'role': 'assistant'
        }
    }

    return response


def _handle_context_completion(knowledge_base_table: KnowledgeBaseTable, request):
    # Used for semantic search.
    query = request.json.get('query')
    # Keyword search.
    keywords = request.json.get('keywords')
    # Metadata search.
    metadata = request.json.get('metadata')
    # Maximum amount of documents to return as context.
    limit = request.json.get('limit', DEFAULT_CONTEXT_DOCUMENT_LIMIT)

    # Use default distance function & column names for ID, content, & metadata, to keep things simple.
    hybrid_search_df = knowledge_base_table.hybrid_search(
        query,
        keywords=keywords,
        metadata=metadata
    )

    num_documents = len(hybrid_search_df.index)
    context_documents = []
    for i in range(limit):
        if i >= num_documents:
            break
        row = hybrid_search_df.iloc[i]
        context_documents.append({
            'id': row['id'],
            'content': row['content'],
            'rank': row['rank']
        })

    return {
        'documents': context_documents
    }


@ns_conf.route('/<project_name>/knowledge_bases/<knowledge_base_name>/completions')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('knowledge_base_name', 'Name of the knowledge_base')
class KnowledgeBaseCompletions(Resource):
    @ns_conf.doc('knowledge_base_completions')
    @api_endpoint_metrics('POST', '/knowledge_bases/knowledge_base/completions')
    def post(self, project_name, knowledge_base_name):
        """
        Add support for LLM generation on the response from knowledge base. Default completion type is 'chat' unless specified.
        """
        if request.json.get('query') is None:
            # "query" is used for semantic search for both completion types.
            logger.error('Missing parameter "query" in POST body')
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "query" parameter in POST body'
            )

        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            logger.error("Project not found, please check the project name exists")
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        session = SessionController()
        # Check if knowledge base exists
        table = session.kb_controller.get_table(knowledge_base_name, project.id)
        if table is None:
            logger.error("Knowledge Base not found, please check the knowledge base name exists")
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

        completion_type = request.json.get('type', 'chat')
        if completion_type == 'context':
            return _handle_context_completion(table, request)
        if completion_type == 'chat':
            return _handle_chat_completion(table, request)
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Invalid parameter',
            f'Completion type must be one of: "context", "chat". Received {completion_type}'
        )



import json
import traceback
from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.api.http.namespaces.configs.tabs import ns_conf
from mindsdb.utilities import log
from mindsdb.api.http.utils import http_error
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.interfaces.tabs.tabs_controller import tabs_controller, get_storage, TABS_FILENAME


logger = log.getLogger(__name__)


def _is_request_valid() -> bool:
    """check if request body contains all (and only) required fields

    Returns:
        bool: True if all required data in the request
    """
    try:
        data = request.json
    except Exception:
        return False
    if (
        isinstance(data, dict) is False
        or len(data.keys()) == 0
        or len(set(data.keys()) - {'index', 'name', 'content'}) != 0
    ):
        return False
    return True


@ns_conf.route('/')
class Tabs(Resource):
    @ns_conf.doc('get_tabs')
    @api_endpoint_metrics('GET', '/tabs')
    def get(self):
        mode = request.args.get('mode')

        if mode == 'new':
            return tabs_controller.get_all(), 200
        else:
            # deprecated
            storage = get_storage()
            tabs = None
            try:
                raw_data = storage.file_get(TABS_FILENAME)
                tabs = json.loads(raw_data)
            except Exception as e:
                logger.warning("unable to get tabs data - %s", e)
                return {}, 200
            return tabs, 200

    @ns_conf.doc('save_tab')
    @api_endpoint_metrics('POST', '/tabs')
    def post(self):
        mode = request.args.get('mode')

        if mode == 'new':
            if _is_request_valid() is False:
                return http_error(400, 'Error', 'Invalid parameters')
            data = request.json
            tab_meta = tabs_controller.add(**data)
            tabs_meta = tabs_controller._get_tabs_meta()
            return {
                'tab_meta': tab_meta,
                'tabs_meta': tabs_meta
            }, 200
        else:
            # deprecated
            storage = get_storage()
            try:
                tabs = request.json
                b_types = json.dumps(tabs).encode("utf-8")
                storage.file_set(TABS_FILENAME, b_types)
            except Exception as e:
                logger.error("unable to store tabs data - %s", e)
                logger.error(traceback.format_exc())
                return http_error(
                    HTTPStatus.INTERNAL_SERVER_ERROR,
                    "Can't save tabs",
                    'something went wrong during tabs saving'
                )

            return '', 200


@ns_conf.route("/<tab_id>")
@ns_conf.param("tab_id", "id of tab")
class Tab(Resource):
    @ns_conf.doc("get_tab")
    @api_endpoint_metrics('GET', '/tabs/tab')
    def get(self, tab_id: int):
        try:
            tab_data = tabs_controller.get(int(tab_id))
        except EntityNotExistsError:
            return http_error(404, 'Error', 'The tab does not exist')

        return tab_data, 200

    @ns_conf.doc("put_tab")
    @api_endpoint_metrics('PUT', '/tabs/tab')
    def put(self, tab_id: int):
        if _is_request_valid() is False:
            return http_error(400, 'Error', 'Invalid parameters')
        data = request.json
        try:
            tab_meta = tabs_controller.modify(int(tab_id), **data)
        except EntityNotExistsError:
            return http_error(404, 'Error', 'The tab does not exist')

        tabs_meta = tabs_controller._get_tabs_meta()

        return {
            'tab_meta': tab_meta,
            'tabs_meta': tabs_meta
        }, 200

    @ns_conf.doc("delete_tab")
    @api_endpoint_metrics('DELETE', '/tabs/tab')
    def delete(self, tab_id: int):
        try:
            tabs_controller.delete(int(tab_id))
        except EntityNotExistsError:
            return http_error(404, 'Error', 'The tab does not exist')
        return '', 200



"""json_storage

Revision ID: cada7d2be947
Revises: 3d5e70105df7
Create Date: 2022-09-29 15:52:32.695026

"""
import json

from alembic import op
import sqlalchemy as sa

import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.storage.fs import RESOURCE_GROUP


# revision identifiers, used by Alembic.
revision = 'cada7d2be947'
down_revision = '3d5e70105df7'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'json_storage',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('resource_group', sa.String(), nullable=True),
        sa.Column('resource_id', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=True),
        sa.Column('content', sa.JSON(), nullable=True),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    predictors = conn.execute(sa.text('''
        select id, json_ai from predictor
    ''')).fetchall()

    for row in predictors:
        try:
            jai = json.loads(row['json_ai'])
        except Exception:
            continue

        if jai is None or len(jai) == 0:
            continue

        record = db.JsonStorage(
            resource_group=RESOURCE_GROUP.PREDICTOR,
            resource_id=row['id'],
            name='json_ai',
            content=jai,
            company_id=None
        )
        session.add(record)

    session.commit()

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('json_ai')


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('json_ai', sa.JSON, nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    jsons = conn.execute(sa.text('''
        select resource_id, name, content
        from json_storage
        where resource_group = 'predictor' and name = 'json_ai'
    ''')).fetchall()

    for row in jsons:
        predicrtor_record = (
            session.query(db.Predictor)
            .filter_by(company_id=None, id=row['resource_id'])
            .first()
        )
        if predicrtor_record is None:
            continue
        predicrtor_record.json_ai = row['content']

    session.commit()

    op.drop_table('json_storage')



import os
import json
import pandas as pd
from typing import Text, Optional, Dict

from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.libs.llm.utils import ft_jsonl_validation, ft_formatter
from mindsdb.integrations.handlers.openai_handler.openai_handler import OpenAIHandler
from mindsdb.integrations.handlers.anyscale_endpoints_handler.settings import anyscale_handler_config
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class AnyscaleEndpointsHandler(OpenAIHandler):
    """
    This handler handles connection and inference with the Anyscale Endpoints API.
    """

    name = 'anyscale_endpoints'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_models = []
        self.chat_completion_models = []
        self.supported_ft_models = []
        self.default_model = anyscale_handler_config.DEFAULT_MODEL
        self.api_base = anyscale_handler_config.ANYSCALE_API_BASE
        self.default_mode = anyscale_handler_config.DEFAULT_MODE
        self.supported_modes = anyscale_handler_config.SUPPORTED_MODES
        self.rate_limit = anyscale_handler_config.RATE_LIMIT
        self.max_batch_size = anyscale_handler_config.MAX_BATCH_SIZE
        self.default_max_tokens = anyscale_handler_config.DEFAULT_MAX_TOKENS

    def create_engine(self, connection_args: Dict) -> None:
        """
        Validate the Anyscale Endpoints credentials on engine creation.

        Args:
            connection_args (Dict): Connection arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """

        connection_args = {k.lower(): v for k, v in connection_args.items()}
        api_key = connection_args.get('anyscale_endpoints_api_key')
        if api_key is not None:
            org = connection_args.get('api_organization')
            api_base = connection_args.get('api_base') or os.environ.get('ANYSCALE_API_BASE', anyscale_handler_config.ANYSCALE_API_BASE)
            client = self._get_client(api_key=api_key, base_url=api_base, org=org)
            OpenAIHandler._check_client_connection(client)

    @staticmethod
    def create_validation(target: Text, args: Optional[Dict] = None, **kwargs: Optional[Dict]) -> None:
        """
        Validate the Anyscale Endpoints credentials on model creation.

        Args:
            target (Text): Target column, not required for LLMs.
            args (Dict): Handler arguments.
            kwargs (Dict): Handler keyword arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """

        if 'using' not in args:
            raise Exception(
                "Anyscale Endpoints engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args['using']

        engine_storage = kwargs['handler_storage']
        connection_args = engine_storage.get_connection_args()
        api_key = get_api_key('anyscale_endpoints', args, engine_storage=engine_storage)
        api_base = connection_args.get('api_base') or args.get('api_base') or os.environ.get('ANYSCALE_API_BASE', anyscale_handler_config.ANYSCALE_API_BASE)

        client = OpenAIHandler._get_client(api_key=api_key, base_url=api_base)
        OpenAIHandler._check_client_connection(client)

    def create(self, target: Text, args: Optional[Dict] = None, **kwargs: Optional[Dict]) -> None:
        """
        Create a model via an engine.

        Args:
            target (Text): Target column.
            args (Dict): Model arguments.
            kwargs (Dict): Other arguments.

        Returns:
            None
        """

        # Set the base and fine-tuned models and call the parent method
        self._set_models_from_args(args)
        super().create(target, args, **kwargs)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Make a prediction using a model.

        Args:
            df (pd.DataFrame): Input data.
            args (Dict): Handler arguments.

        Returns:
            pd.DataFrame: Predicted data
        """

        # Set the base and fine-tuned models and call the parent method
        self._set_models_from_args(args)
        return super().predict(df, args)

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Fine-tune a supported model.

        Args:
            df (pd.DataFrame): Input data.
            args (Dict): Handler arguments.

        Returns:
            None
        """

        using_args = args.get('using', {})
        self._set_models(using_args)
        super().finetune(df, args)

        # Rewrite chat_completion_models to include the newly fine-tuned model
        args = self.model_storage.json_get('args')
        args['chat_completion_models'] = list(self.chat_completion_models) + [args['model_name']]
        self.model_storage.json_set('args', args)

    def describe(self, attribute: Optional[Text] = None) -> pd.DataFrame:
        """
        Describe a model or its metadata.

        Args:
            attribute (Text): Attribute to describe.

        Returns:
            pd.DataFrame: Model or metadata description.
        """

        args = self.model_storage.json_get('args')

        # Remove keys from args to display
        for arg in ('api_key', 'openai_api_key'):
            if arg in args:
                del args[arg]

        if attribute == 'args':
            return pd.DataFrame(args.items(), columns=['key', 'value'])
        elif attribute == 'metadata':
            # The URL is used as some models require completing a form to access their artifacts
            model_name = args.get('model_name', self.default_model)
            model_card_url = 'https://huggingface.co/' + model_name
            return pd.DataFrame({'model_name': [model_name], 'model_card': [model_card_url]})
        else:
            tables = ['args', 'metadata']
            return pd.DataFrame(tables, columns=['tables'])

    def _set_models(self, args: Dict) -> None:
        """
        Set the base and fine-tuned models.

        Args:
            args (Dict): Model arguments.

        Returns:
            None
        """

        api_key = get_api_key('anyscale_endpoints', args, self.engine_storage)
        client = OpenAIHandler._get_client(api_key=api_key, base_url=self.api_base)
        self.all_models = [m.id for m in client.models.list()]
        self.chat_completion_models = [m.id for m in client.models.list() if m.rayllm_metadata['engine_config']['model_type'] == 'text-generation']  # noqa
        # Set base models compatible with fine-tuning
        self.supported_ft_models = self.chat_completion_models

    def _set_models_from_args(self, args: Dict) -> None:
        """
        Set the base and fine-tuned models from the arguments, if specified. Otherwise, use the default list.

        Args:
            args (Dict): Model arguments.

        Returns:
            None
        """

        self._set_models(args.get('using', {}))

        # Update the models if they are specified in the arguments
        model_args = self.model_storage.json_get('args')
        if model_args and 'chat_completion_models' in model_args:
            self.chat_completion_models = model_args.get('chat_completion_models')

    @staticmethod
    def _prepare_ft_jsonl(df, temp_storage_path: Text, temp_filename: Text, _, test_size: Optional[float] = 0.2) -> Dict:
        """
        Prepare the data for fine-tuning.

        Args:
            df (pd.DataFrame): Input data.
            temp_storage_path (Text): Temporary storage path.
            temp_filename (Text): Temporary filename.
            _: Unused.
            test_size (float): Test size.

        Returns:
            dict: File names mapped to the prepared data.
        """

        # 1. Format data
        chats = ft_formatter(df)

        # 2. Split chats in training and validation subsets
        series = pd.Series(chats)
        if len(series) < anyscale_handler_config.MIN_FT_DATASET_LEN:
            raise Exception(f"Dataset is too small to finetune. Please include at least {anyscale_handler_config.MIN_FT_DATASET_LEN} samples (complete chats).")
        val_size = max(anyscale_handler_config.MIN_FT_VAL_LEN, int(len(series) * test_size))  # at least as many samples as required by Anyscale
        train = series.iloc[:-val_size]
        val = series.iloc[-val_size:]

        # 3. Write as JSONL files
        file_names = {
            'train': f'{temp_filename}_prepared_train.jsonl',
            'val': f'{temp_filename}_prepared_valid.jsonl',
        }
        train.to_json(os.path.join(temp_storage_path, file_names['train']), orient='records', lines=True)
        val.to_json(os.path.join(temp_storage_path, file_names['val']), orient='records', lines=True)

        # 4. Validate and return
        with open(os.path.join(temp_storage_path, file_names['train']), 'r', encoding='utf-8') as f:
            ft_jsonl_validation([json.loads(line) for line in f])

        with open(os.path.join(temp_storage_path, file_names['val']), 'r', encoding='utf-8') as f:
            ft_jsonl_validation([json.loads(line) for line in f])

        return file_names

    def _get_ft_model_type(self, model_name: Text) -> Text:
        """
        Get the fine-tuning model type.

        Args:
            model_name (Text): Model name.

        Returns:
            Text: Model type.
        """

        for base_model in self.chat_completion_models:
            if base_model.lower() in model_name.lower():
                return base_model
        logger.warning(f'Cannot recognize model {model_name}. Finetuning may fail.')
        return model_name.lower()

    @staticmethod
    def _add_extra_ft_params(ft_params: Dict, using_args: Dict) -> Dict:
        """
        Add extra fine-tuning parameters.

        Args:
            ft_params (Dict): Fine-tuning parameters.
            using_args (Dict): Model arguments.

        Returns:
            Dict: Fine-tuning parameters with extra parameters.
        """

        hyperparameters = {}
        # Populate separately because keys with `None` break the API
        for key in ('n_epochs', 'context_length'):
            if using_args.get(key, None):
                hyperparameters[key] = using_args[key]
        if hyperparameters:
            return {**ft_params, **{'hyperparameters': hyperparameters}}
        else:
            return ft_params



import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.handlers.vertex_handler.vertex_client import VertexClient
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class VertexHandler(BaseMLEngine):
    """Handler for the Vertex Google AI cloud API"""

    name = "Vertex"

    def create(self, target, args=None, **kwargs):
        """Logs in to Vertex and deploy a pre-trained model to an endpoint.

        If the endpoint already exists for the model, we do nothing.

        If the endpoint does not exist, we create it and deploy the model to it.
        The runtime for this is long, it took 15 minutes for a small model.
        """
        assert "using" in args, "Must provide USING arguments for this handler"
        args = args["using"]

        model_name = args.pop("model_name")
        custom_model = args.pop("custom_model", False)

        # get credentials from engine
        credentials_url, credentials_file, credentials_json = self._get_credentials_from_engine()

        # get vertex args from handler then update args from model
        vertex_args = self.engine_storage.json_get('args')
        vertex_args.update(args)

        vertex = VertexClient(vertex_args, credentials_url, credentials_file, credentials_json)

        model = vertex.get_model_by_display_name(model_name)
        if not model:
            raise Exception(f"Vertex model {model_name} not found")
        endpoint_name = model_name + "_endpoint"
        if vertex.get_endpoint_by_display_name(endpoint_name):
            logger.info(f"Endpoint {endpoint_name} already exists, skipping deployment")
        else:
            logger.info(f"Starting deployment at {endpoint_name}")
            endpoint = vertex.deploy_model(model)
            endpoint.display_name = endpoint_name
            endpoint.update()
            logger.info(f"Endpoint {endpoint_name} deployed")

        predict_args = {}
        predict_args["target"] = target
        predict_args["endpoint_name"] = endpoint_name
        predict_args["custom_model"] = custom_model
        self.model_storage.json_set("predict_args", predict_args)
        self.model_storage.json_set("vertex_args", vertex_args)

    def predict(self, df, args=None):
        """Predict using the deployed model by calling the endpoint."""

        if "__mindsdb_row_id" in df.columns:
            df.drop("__mindsdb_row_id", axis=1, inplace=True)  # TODO is this required?

        predict_args = self.model_storage.json_get("predict_args")
        vertex_args = self.model_storage.json_get("vertex_args")

        # get credentials from engine
        credentials_url, credentials_file, credentials_json = self._get_credentials_from_engine()

        vertex = VertexClient(vertex_args, credentials_url, credentials_file, credentials_json)
        results = vertex.predict_from_df(predict_args["endpoint_name"], df, custom_model=predict_args["custom_model"])

        if predict_args["custom_model"]:
            return pd.DataFrame(results.predictions, columns=[predict_args["target"]])
        else:
            return pd.DataFrame(results.predictions)

    def create_engine(self, connection_args):
        # check if one of credentials_url, credentials_file, or credentials_json is provided
        if 'service_account_key_url' not in connection_args and 'service_account_key_file' not in connection_args and 'service_account_key_json' not in connection_args:
            raise KeyError('Either service_account_key_url, service_account_key_file, or service_account_key_json must be provided')

        self.engine_storage.json_set('args', connection_args)

    def _get_credentials_from_engine(self):
        engine_args = self.engine_storage.json_get('args')

        return engine_args.get('service_account_key_url'), engine_args.get('service_account_key_file'), engine_args.get('service_account_key_json')



from typing import List
from mindsdb.api.executor.sql_query.result_set import ResultSet


class ExecuteAnswer:
    def __init__(
        self,
        data: ResultSet = None,
        state_track: List[List] = None,
        error_code: int = None,
        error_message: str = None,
    ):
        self.data = data
        self.state_track = state_track
        self.error_code = error_code
        self.error_message = error_message



import difflib
from typing import Any

from pydantic import BaseModel, model_validator


class ChromaHandlerConfig(BaseModel):
    """
    Configuration for VectorStoreHandler.
    """

    vector_store: str
    persist_directory: str = None
    host: str = None
    port: str = None
    password: str = None

    class Config:
        extra = "forbid"

    @model_validator(mode="before")
    @classmethod
    def check_param_typos(cls, values: Any) -> Any:
        """Check if there are any typos in the parameters."""

        expected_params = cls.model_fields.keys()
        for key in values.keys():
            if key not in expected_params:
                close_matches = difflib.get_close_matches(
                    key, expected_params, cutoff=0.4
                )
                if close_matches:
                    raise ValueError(
                        f"Unexpected parameter '{key}'. Did you mean '{close_matches[0]}'?"
                    )
                else:
                    raise ValueError(f"Unexpected parameter '{key}'.")
        return values

    @model_validator(mode="before")
    @classmethod
    def check_config(cls, values: Any) -> Any:
        """Check if config is valid."""

        vector_store = values.get("vector_store")
        host = values.get("host")
        port = values.get("port")
        persist_directory = values.get("persist_directory")

        if bool(port) != bool(host) or (host and persist_directory):
            raise ValueError(
                f"For {vector_store} handler - host and port must be provided together. "
                f"Additionally, if host and port are provided, persist_directory should not be provided."
            )

        if persist_directory and (host or port):
            raise ValueError(
                f"For {vector_store} handler - if persistence_folder is provided, "
                f"host, port should not be provided."
            )

        return values



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    app_id={
        "type": ARG_TYPE.PWD,
        "description": "App ID of symbl",
        "required": True,
        "label": "app_id",
    },
    app_secret={
        "type": ARG_TYPE.PWD,
        "description": "App Secret of symbl",
        "required": True,
        "label": "app_secret",
        "secret": True
    }
)

connection_args_example = OrderedDict(
    app_id="app_id",
    app_secret="app_secret"
)



def table_names(connection):
    query = "INFO for DB"
    dict_1 = connection.query(query)
    dict_2 = dict_1['tb']
    tables = list(dict_2.keys())
    return tables


def column_info(connection, table):
    query = "INFO FOR TABLE " + table
    dict_1 = connection.query(query)
    dict_2 = dict_1['fd']
    columns = list(dict_2.keys())
    types = []

    for value in dict_2.values():
        a = value.split('TYPE ', 1)[1]
        type = a.split()[0]
        types.append(type)
    return columns, types



from typing import Any, Dict, Text

from hdbcli import dbapi
from hdbcli.dbapi import Error, ProgrammingError
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from pandas import DataFrame
import sqlalchemy_hana.dialect as hana_dialect

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class HanaHandler(DatabaseHandler):
    """
    This handler handles the connection and execution of SQL statements on SAP HANA.
    """

    name = 'hana'

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the SAP HANA database.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        """
        Closes the connection when the handler instance is deleted.
        """
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> dbapi.Connection:
        """
        Establishes a connection to the SAP HANA database.

        Raises:
            ValueError: If the expected connection parameters are not provided.
            hdbcli.dbapi.Error: If an error occurs while connecting to the SAP HANA database.

        Returns:
            hdbcli.dbapi.Connection: A connection object to the SAP HANA database.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters.
        if not all(key in self.connection_data for key in ['address', 'port', 'user', 'password']):
            raise ValueError('Required parameters (address, port, user, password) must be provided.')

        config = {
            'address': self.connection_data['address'],
            'port': self.connection_data['port'],
            'user': self.connection_data['user'],
            'password': self.connection_data['password'],
        }

        # Optional connection parameters.
        if 'database' in self.connection_data:
            config['databaseName'] = self.connection_data['database']

        if 'schema' in self.connection_data:
            config['currentSchema'] = self.connection_data['schema']

        if 'encrypt' in self.connection_data:
            config['encrypt'] = self.connection_data['encrypt']

        try:
            self.connection = dbapi.connect(
                **config
            )
            self.is_connected = True
            return self.connection
        except Error as known_error:
            logger.error(f'Error connecting to SAP HANA, {known_error}!')
            raise
        except Exception as unknown_error:
            logger.error(f'Unknown error connecting to Teradata, {unknown_error}!')
            raise

    def disconnect(self) -> None:
        """
        Closes the connection to the SAP HANA database if it's currently open.
        """
        if self.is_connected is True:
            self.connection.close()
            self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the SAP HANA database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('SELECT 1 FROM SYS.DUMMY')
            response.success = True
        except (Error, ProgrammingError, ValueError) as known_error:
            logger.error(f'Connection check to SAP HANA failed, {known_error}!')
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f'Connection check to SAP HANA failed due to an unknown error, {unknown_error}!')
            response.error_message = str(unknown_error)

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a native SQL query on the SAP HANA database and returns the result.

        Args:
            query (Text): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if not cur.description:
                    response = Response(RESPONSE_TYPE.OK)
                else:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                connection.commit()
            except ProgrammingError as programming_error:
                logger.error(f'Error running query: {query} on {self.address}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(programming_error)
                )
                connection.rollback()
            except Exception as unknown_error:
                logger.error(f'Unknown error running query: {query} on {self.address}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(unknown_error)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode on the SAP HANA database and retrieves the data (if any).

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender(hana_dialect.HANAHDBCLIDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables in the SAP HANA database.

        Returns:
            Response: A response object containing a list of tables in the SAP HANA database.
        """
        query = """
            SELECT SCHEMA_NAME,
                   TABLE_NAME,
                   'BASE TABLE' AS TABLE_TYPE
            FROM
                SYS.TABLES
            WHERE IS_SYSTEM_TABLE = 'FALSE'
              AND IS_USER_DEFINED_TYPE = 'FALSE'
              AND IS_TEMPORARY = 'FALSE'

            UNION

            SELECT SCHEMA_NAME,
                   VIEW_NAME AS TABLE_NAME,
                   'VIEW' AS TABLE_TYPE
            FROM
                SYS.VIEWS
            WHERE SCHEMA_NAME <> 'SYS'
              AND SCHEMA_NAME NOT LIKE '_SYS%'
        """
        return self.native_query(query)

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column details for a specified table in the SAP HANA database.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"""
            SELECT COLUMN_NAME AS Field,
                DATA_TYPE_NAME AS Type
            FROM SYS.TABLE_COLUMNS
            WHERE TABLE_NAME = '{table_name}'

            UNION ALL

            SELECT COLUMN_NAME AS Field,
                DATA_TYPE_NAME AS Type
            FROM SYS.VIEW_COLUMNS
            WHERE VIEW_NAME = '{table_name}'
        """
        return self.native_query(query)



import re
import time
import threading

from bson import ObjectId
from mindsdb_sql_parser.ast.base import ASTNode
import pandas as pd
import pymongo
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError, OperationFailure, ConfigurationError, InvalidURI
from typing import Text, List, Dict, Any, Union

from mindsdb.api.mongo.utilities.mongodb_query import MongoQuery
from mindsdb.api.mongo.utilities.mongodb_parser import MongodbParser
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log
from .utils.mongodb_render import MongodbRender


logger = log.getLogger(__name__)


class MongoDBHandler(DatabaseHandler):
    """
    This handler handles the connection and execution of SQL statements on MongoDB.
    """

    _SUBSCRIBE_SLEEP_INTERVAL = 0.5
    name = 'mongodb'

    def __init__(self, name: Text, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            kwargs: Arbitrary keyword arguments including the connection data.
        """
        super().__init__(name)
        connection_data = kwargs['connection_data']
        self.host = connection_data.get("host")
        self.port = int(connection_data.get("port") or 27017)
        self.user = connection_data.get("username")
        self.password = connection_data.get("password")
        self.database = connection_data.get('database')
        self.flatten_level = connection_data.get('flatten_level', 0)

        self.connection = None
        self.is_connected = False

    def __del__(self) -> None:
        """
        Closes the connection when the handler instance is deleted.
        """
        if self.is_connected:
            self.disconnect()

    def connect(self) -> MongoClient:
        """
        Establishes a connection to the MongoDB host.

        Raises:
            ValueError: If the expected connection parameters are not provided.

        Returns:
            pymongo.MongoClient: A connection object to the MongoDB host.
        """
        kwargs = {}
        if isinstance(self.user, str) and len(self.user) > 0:
            kwargs['username'] = self.user

        if isinstance(self.password, str) and len(self.password) > 0:
            kwargs['password'] = self.password

        if re.match(r'/?.*tls=true', self.host.lower()):
            kwargs['tls'] = True

        if re.match(r'/?.*tls=false', self.host.lower()):
            kwargs['tls'] = False

        try:
            connection = MongoClient(
                self.host,
                port=self.port,
                **kwargs
            )
        except InvalidURI as invalid_uri_error:
            logger.error(f'Invalid URI provided for MongoDB connection: {invalid_uri_error}!')
            raise
        except ConfigurationError as config_error:
            logger.error(f'Configuration error connecting to MongoDB: {config_error}!')
            raise
        except Exception as unknown_error:
            logger.error(f'Unknown error connecting to MongoDB: {unknown_error}!')
            raise

        # Get the database name from the connection if it's not provided.
        if self.database is None:
            self.database = connection.get_database().name

        self.is_connected = True
        self.connection = connection
        return self.connection

    def subscribe(self, stop_event: threading.Event, callback: callable, table_name: Text, columns: List = None, **kwargs: Any) -> None:
        """
        Subscribes to changes in a MongoDB collection and calls the provided callback function when changes occur.

        Args:
            stop_event (threading.Event): An event object to stop the subscription.
            callback (callable): The callback function to call when changes occur.
            table_name (Text): The name of the collection to subscribe to.
            columns (List): A list of columns to monitor for changes.
            kwargs: Arbitrary keyword arguments.
        """
        con = self.connect()
        cur = con[self.database][table_name].watch()

        while True:
            if stop_event.is_set():
                cur.close()
                return

            res = cur.try_next()
            if res is None:
                time.sleep(self._SUBSCRIBE_SLEEP_INTERVAL)
                continue

            _id = res['documentKey']['_id']
            if res['operationType'] == 'insert':
                if columns is not None:
                    updated_columns = set(res['fullDocument'].keys())
                    if not set(columns) & set(updated_columns):
                        # Do nothing.
                        continue

                callback(row=res['fullDocument'], key={'_id': _id})

            if res['operationType'] == 'update':
                if columns is not None:
                    updated_columns = set(res['updateDescription']['updatedFields'].keys())
                    if not set(columns) & set(updated_columns):
                        # Do nothing.
                        continue

                # Get the full document.
                full_doc = con[self.database][table_name].find_one(res['documentKey'])
                callback(row=full_doc, key={'_id': _id})

    def disconnect(self) -> None:
        """
        Closes the connection to the MongoDB host if it's currently open.
        """
        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the MongoDB host.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            con = self.connect()
            con.server_info()

            # Check if the database exists.
            if self.database not in con.list_database_names():
                raise ValueError(f'Database {self.database} not found!')

            response.success = True
        except (InvalidURI, ServerSelectionTimeoutError, OperationFailure, ConfigurationError, ValueError) as known_error:
            logger.error(f'Error connecting to MongoDB {self.database}, {known_error}!')
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f'Unknown error connecting to MongoDB {self.database}, {unknown_error}!')
            response.error_message = str(unknown_error)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: Union[Text, Dict, MongoQuery]) -> Response:
        """
        Executes a SQL query on the MongoDB host and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        if isinstance(query, str):
            query = MongodbParser().from_string(query)

        if isinstance(query, dict):
            # Fallback for the previous API.
            mquery = MongoQuery(query['collection'])

            for c in query['call']:
                mquery.add_step({
                    'method': c['method'],
                    'args': c['args']
                })

            query = mquery

        collection = query.collection
        database = self.database

        con = self.connect()

        # Check if the collection exists.
        if collection not in con[database].list_collection_names():
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message=f'Collection {collection} not found in database {database}!'
            )

        try:
            cursor = con[database][collection]

            for step in query.pipeline:
                fnc = getattr(cursor, step['method'])
                cursor = fnc(*step['args'])

            result = []
            if not isinstance(cursor, pymongo.results.UpdateResult):
                for row in cursor:
                    result.append(self.flatten(row, level=self.flatten_level))

            else:
                return Response(RESPONSE_TYPE.OK)

            if len(result) > 0:
                df = pd.DataFrame(result)
            else:
                columns = list(self.get_columns(collection).data_frame.Field)
                df = pd.DataFrame([], columns=columns)

            response = Response(
                RESPONSE_TYPE.TABLE,
                df
            )
        except Exception as e:
            logger.error(f'Error running query: {query} on {self.database}.{collection}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        return response

    def flatten(self, row: Dict, level: int = 0) -> Dict:
        """
        Flattens a nested dictionary to a single level.

        Args:
            row (Dict): The dictionary to flatten.
            level (int): The number of levels to flatten. If 0, the entire dictionary is flattened.

        Returns:
            Dict: The flattened dictionary.
        """
        add = {}
        del_keys = []
        edit_keys = {}

        for k, v in row.items():
            # Convert ObjectId to string.
            if isinstance(v, ObjectId):
                edit_keys[k] = str(v)
            if level > 0:
                if isinstance(v, dict):
                    for k2, v2 in self.flatten(v, level=level - 1).items():
                        add[f'{k}.{k2}'] = v2
                    del_keys.append(k)

        if add:
            row.update(add)
        for key in del_keys:
            del row[key]
        if edit_keys:
            row.update(edit_keys)

        return row

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode on the MongoDB host and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = MongodbRender()
        mquery = renderer.to_mongo_query(query)
        return self.native_query(mquery)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables (collections) in the MongoDB host.

        Returns:
            Response: A response object containing a list of tables (collections) in the MongoDB host.
        """
        con = self.connect()
        collections = con[self.database].list_collection_names()
        collections_ar = [
            [i] for i in collections
        ]
        df = pd.DataFrame(collections_ar, columns=['table_name'])

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column (field) details for a specified table (collection) in the MongoDB host.
        The first record in the collection is used to determine the column details.

        Args:
            table_name (Text): The name of the table (collection) for which to retrieve column (field) information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        con = self.connect()
        record = con[self.database][table_name].find_one()

        data = []
        if record is not None:
            record = self.flatten(record)

            for k, v in record.items():
                data.append([k, type(v).__name__])

        df = pd.DataFrame(data, columns=['Field', 'Type'])

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )
        return response



from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import CAPABILITIES


class ClentCapabilities():
    _capabilities = 0

    def __init__(self, capabilities):
        self._capabilities = capabilities

    def has(self, cap):
        return cap & self._capabilities > 0

    def to_dict(self):
        funcs = [func for func in dir(self) if func.upper() == func]
        res = {}
        for f in funcs:
            res[f] = getattr(self, f)
        return res

    @property
    def LONG_PASSWORD(self):
        return self.has(CAPABILITIES.CLIENT_LONG_PASSWORD)

    @property
    def FOUND_ROWS(self):
        return self.has(CAPABILITIES.CLIENT_FOUND_ROWS)

    @property
    def LONG_FLAG(self):
        return self.has(CAPABILITIES.CLIENT_LONG_FLAG)

    @property
    def CONNECT_WITH_DB(self):
        return self.has(CAPABILITIES.CLIENT_CONNECT_WITH_DB)

    @property
    def NO_SCHEMA(self):
        return self.has(CAPABILITIES.CLIENT_NO_SCHEMA)

    @property
    def COMPRESS(self):
        return self.has(CAPABILITIES.CLIENT_COMPRESS)

    @property
    def ODBC(self):
        return self.has(CAPABILITIES.CLIENT_ODBC)

    @property
    def LOCAL_FILES(self):
        return self.has(CAPABILITIES.CLIENT_LOCAL_FILES)

    @property
    def IGNORE_SPACE(self):
        return self.has(CAPABILITIES.CLIENT_IGNORE_SPACE)

    @property
    def PROTOCOL_41(self):
        return self.has(CAPABILITIES.CLIENT_PROTOCOL_41)

    @property
    def INTERACTIVE(self):
        return self.has(CAPABILITIES.CLIENT_INTERACTIVE)

    @property
    def SSL(self):
        return self.has(CAPABILITIES.CLIENT_SSL)

    @property
    def IGNORE_SIGPIPE(self):
        return self.has(CAPABILITIES.CLIENT_IGNORE_SIGPIPE)

    @property
    def TRANSACTIONS(self):
        return self.has(CAPABILITIES.CLIENT_TRANSACTIONS)

    @property
    def RESERVED(self):
        return self.has(CAPABILITIES.CLIENT_RESERVED)

    @property
    def RESERVED2(self):
        return self.has(CAPABILITIES.CLIENT_RESERVED2)

    @property
    def MULTI_STATEMENTS(self):
        return self.has(CAPABILITIES.CLIENT_MULTI_STATEMENTS)

    @property
    def MULTI_RESULTS(self):
        return self.has(CAPABILITIES.CLIENT_MULTI_RESULTS)

    @property
    def PS_MULTI_RESULTS(self):
        return self.has(CAPABILITIES.CLIENT_PS_MULTI_RESULTS)

    @property
    def PLUGIN_AUTH(self):
        return self.has(CAPABILITIES.CLIENT_PLUGIN_AUTH)

    @property
    def CONNECT_ATTRS(self):
        return self.has(CAPABILITIES.CLIENT_CONNECT_ATTRS)

    @property
    def PLUGIN_AUTH_LENENC_CLIENT_DATA(self):
        return self.has(CAPABILITIES.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA)

    @property
    def CAN_HANDLE_EXPIRED_PASSWORDS(self):
        return self.has(CAPABILITIES.CLIENT_CAN_HANDLE_EXPIRED_PASSWORDS)

    @property
    def SESSION_TRACK(self):
        return self.has(CAPABILITIES.CLIENT_SESSION_TRACK)

    @property
    def DEPRECATE_EOF(self):
        return self.has(CAPABILITIES.CLIENT_DEPRECATE_EOF)

    @property
    def SSL_VERIFY_SERVER_CERT(self):
        return self.has(CAPABILITIES.CLIENT_SSL_VERIFY_SERVER_CERT)

    @property
    def REMEMBER_OPTIONS(self):
        return self.has(CAPABILITIES.CLIENT_REMEMBER_OPTIONS)

    @property
    def SECURE_CONNECTION(self):
        return self.has(CAPABILITIES.CLIENT_SECURE_CONNECTION)



import datetime
from typing import Dict, Iterator, List, Union, Tuple, Optional

from langchain_core.tools import BaseTool
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy import null
import pandas as pd

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.db import Predictor
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.skills.skills_controller import SkillsController
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError

from .constants import ASSISTANT_COLUMN, SUPPORTED_PROVIDERS, PROVIDER_TO_MODELS
from .langchain_agent import get_llm_provider


class AgentsController:
    '''Handles CRUD operations at the database level for Agents'''

    assistant_column = ASSISTANT_COLUMN

    def __init__(
        self,
        project_controller: ProjectController = None,
        skills_controller: SkillsController = None,
        model_controller: ModelController = None
    ):
        if project_controller is None:
            project_controller = ProjectController()
        if skills_controller is None:
            skills_controller = SkillsController()
        if model_controller is None:
            model_controller = ModelController()
        self.project_controller = project_controller
        self.skills_controller = skills_controller
        self.model_controller = model_controller

    def check_model_provider(self, model_name: str, provider: str = None) -> Tuple[dict, str]:
        '''
        Checks if a model exists, and gets the provider of the model.

        The provider is either the provider of the model, or the provider given as an argument.

        Parameters:
            model_name (str): The name of the model
            provider (str): The provider to check

        Returns:
            model (dict): The model object
            provider (str): The provider of the model
        '''
        model = None

        try:
            model_name_no_version, model_version = Predictor.get_name_and_version(model_name)
            model = self.model_controller.get_model(model_name_no_version, version=model_version)
            provider = 'mindsdb' if model.get('provider') is None else model.get('provider')
        except PredictorRecordNotFound:
            if not provider:
                # If provider is not given, get it from the model name
                provider = get_llm_provider({"model_name": model_name})

            elif provider not in SUPPORTED_PROVIDERS and model_name not in PROVIDER_TO_MODELS.get(provider, []):
                raise ValueError(f'Model with name does not exist for provider {provider}: {model_name}')

        return model, provider

    def get_agent(self, agent_name: str, project_name: str = 'mindsdb') -> Optional[db.Agents]:
        '''
        Gets an agent by name.

        Parameters:
            agent_name (str): The name of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (Optional[db.Agents]): The database agent object
        '''

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.name == agent_name,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id,
            db.Agents.deleted_at == null()
        ).first()
        return agent

    def get_agent_by_id(self, id: int, project_name: str = 'mindsdb') -> db.Agents:
        '''
        Gets an agent by id.

        Parameters:
            id (int): The id of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (db.Agents): The database agent object
        '''

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.id == id,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id,
            db.Agents.deleted_at == null()
        ).first()
        return agent

    def get_agents(self, project_name: str) -> List[dict]:
        """
        Gets all agents in a project.

        Parameters:
            project_name (str): The name of the containing project - must exist

        Returns:
            all-agents (List[db.Agents]): List of database agent object
        """

        all_agents = db.Agents.query.filter(
            db.Agents.company_id == ctx.company_id,
            db.Agents.deleted_at == null()
        )

        if project_name is not None:
            project = self.project_controller.get(name=project_name)

            all_agents = all_agents.filter(db.Agents.project_id == project.id)

        return all_agents.all()

    def add_agent(
            self,
            name: str,
            project_name: str,
            model_name: str,
            skills: List[Union[str, dict]],
            provider: str = None,
            params: Dict[str, str] = {}) -> db.Agents:
        '''
        Adds an agent to the database.

        Parameters:
            name (str): The name of the new agent
            project_name (str): The containing project
            model_name (str): The name of the existing ML model the agent will use
            skills (List[Union[str, dict]]): List of existing skill names to add to the new agent, or list of dicts
                 with one of keys is "name", and other is additional parameters for relationship agent<>skill
            provider (str): The provider of the model
            params (Dict[str, str]): Parameters to use when running the agent

        Returns:
            agent (db.Agents): The created agent

        Raises:
            ValueError: Agent with given name already exists, or skill/model with given name does not exist.
        '''
        if project_name is None:
            project_name = 'mindsdb'
        project = self.project_controller.get(name=project_name)

        agent = self.get_agent(name, project_name)

        if agent is not None:
            raise ValueError(f'Agent with name already exists: {name}')

        _, provider = self.check_model_provider(model_name, provider)

        agent = db.Agents(
            name=name,
            project_id=project.id,
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            model_name=model_name,
            provider=provider,
            params=params,
        )

        for skill in skills:
            if isinstance(skill, str):
                skill_name = skill
                parameters = {}
            else:
                parameters = skill.copy()
                skill_name = parameters.pop('name')
            existing_skill = self.skills_controller.get_skill(skill_name, project_name)
            if existing_skill is None:
                db.session.rollback()
                raise ValueError(f'Skill with name does not exist: {skill_name}')
            association = db.AgentSkillsAssociation(
                parameters=parameters,
                agent=agent,
                skill=existing_skill
            )
            db.session.add(association)

        db.session.add(agent)
        db.session.commit()

        return agent

    def update_agent(
            self,
            agent_name: str,
            project_name: str = 'mindsdb',
            name: str = None,
            model_name: str = None,
            skills_to_add: List[Union[str, dict]] = None,
            skills_to_remove: List[str] = None,
            skills_to_rewrite: List[Union[str, dict]] = None,
            provider: str = None,
            params: Dict[str, str] = None):
        '''
        Updates an agent in the database.

        Parameters:
            agent_name (str): The name of the new agent, or existing agent to update
            project_name (str): The containing project
            name (str): The updated name of the agent
            model_name (str): The name of the existing ML model the agent will use
            skills_to_add (List[Union[str, dict]]): List of skill names to add to the agent, or list of dicts
                 with one of keys is "name", and other is additional parameters for relationship agent<>skill
            skills_to_remove (List[str]): List of skill names to remove from the agent
            skills_to_rewrite (List[Union[str, dict]]): new list of skills for the agent
            provider (str): The provider of the model
            params: (Dict[str, str]): Parameters to use when running the agent

        Returns:
            agent (db.Agents): The created or updated agent

        Raises:
            EntityExistsError: if agent with new name already exists
            EntityNotExistsError: if agent with name or skill not found
            ValueError: if conflict in skills list
        '''

        skills_to_add = skills_to_add or []
        skills_to_remove = skills_to_remove or []
        skills_to_rewrite = skills_to_rewrite or []

        if len(skills_to_rewrite) > 0 and (len(skills_to_remove) > 0 or len(skills_to_add) > 0):
            raise ValueError(
                "'skills_to_rewrite' and 'skills_to_add' (or 'skills_to_remove') cannot be used at the same time"
            )

        existing_agent = self.get_agent(agent_name, project_name=project_name)
        if existing_agent is None:
            raise EntityNotExistsError(f'Agent with name not found: {agent_name}')
        is_demo = (existing_agent.params or {}).get('is_demo', False)
        if (
            is_demo and (
                (name is not None and name != agent_name)
                or (model_name or provider)
                or (len(skills_to_add) > 0 or len(skills_to_remove) > 0 or len(skills_to_rewrite) > 0)
                or (isinstance(params, dict) and len(params) > 1 and 'prompt_template' not in params)
            )
        ):
            raise ValueError("It is forbidden to change properties of the demo object")

        if name is not None and name != agent_name:
            # Check to see if updated name already exists
            agent_with_new_name = self.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                raise EntityExistsError(f'Agent with updated name already exists: {name}')
            existing_agent.name = name

        if model_name or provider:
            # check model and provider
            model, provider = self.check_model_provider(model_name, provider)
            # Update model and provider
            existing_agent.model_name = model_name
            existing_agent.provider = provider

        # check that all skills exist
        skill_name_to_record_map = {}
        for skill_meta in (skills_to_add + skills_to_remove + skills_to_rewrite):
            skill_name = skill_meta['name'] if isinstance(skill_meta, dict) else skill_meta
            if skill_name not in skill_name_to_record_map:
                skill_record = self.skills_controller.get_skill(skill_name, project_name)
                if skill_record is None:
                    raise EntityNotExistsError(f'Skill with name does not exist: {skill_name}')
                skill_name_to_record_map[skill_name] = skill_record

        if len(skills_to_add) > 0 or len(skills_to_remove) > 0:
            skills_to_add = [{'name': x} if isinstance(x, str) else x for x in skills_to_add]
            skills_to_add_names = [x['name'] for x in skills_to_add]

            # there are no intersection between lists
            if not set(skills_to_add_names).isdisjoint(set(skills_to_remove)):
                raise ValueError('Conflict between skills to add and skills to remove.')

            existing_agent_skills_names = [rel.skill.name for rel in existing_agent.skills_relationships]

            # remove skills
            for skill_name in skills_to_remove:
                for rel in existing_agent.skills_relationships:
                    if rel.skill.name == skill_name:
                        db.session.delete(rel)

            # add skills
            for skill_name in (set(skills_to_add_names) - set(existing_agent_skills_names)):
                skill_parameters = next(x for x in skills_to_add if x['name'] == skill_name).copy()
                del skill_parameters['name']
                association = db.AgentSkillsAssociation(
                    parameters=skill_parameters,
                    agent=existing_agent,
                    skill=skill_name_to_record_map[skill_name]
                )
                db.session.add(association)
        elif len(skills_to_rewrite) > 0:
            skill_name_to_parameters = {x['name']: {
                k: v for k, v in x.items() if k != 'name'
            } for x in skills_to_rewrite}
            existing_skill_names = set()
            for rel in existing_agent.skills_relationships:
                if rel.skill.name not in skill_name_to_parameters:
                    db.session.delete(rel)
                else:
                    existing_skill_names.add(rel.skill.name)
                    rel.parameters = skill_name_to_parameters[rel.skill.name]
                    flag_modified(rel, 'parameters')
            for new_skill_name in (set(skill_name_to_parameters) - existing_skill_names):
                association = db.AgentSkillsAssociation(
                    parameters=skill_name_to_parameters[new_skill_name],
                    agent=existing_agent,
                    skill=skill_name_to_record_map[new_skill_name]
                )
                db.session.add(association)

        if params is not None:
            # Merge params on update
            existing_params = existing_agent.params or {}
            existing_params.update(params)
            # Remove None values entirely.
            params = {k: v for k, v in existing_params.items() if v is not None}
            existing_agent.params = params
            # Some versions of SQL Alchemy won't handle JSON updates correctly without this.
            # See: https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.attributes.flag_modified
            flag_modified(existing_agent, 'params')
        db.session.commit()

        return existing_agent

    def delete_agent(self, agent_name: str, project_name: str = 'mindsdb'):
        '''
        Deletes an agent by name.

        Parameters:
            agent_name (str): The name of the agent to delete
            project_name (str): The name of the containing project

        Raises:
            ValueError: Agent does not exist.
        '''

        agent = self.get_agent(agent_name, project_name)
        if agent is None:
            raise ValueError(f'Agent with name does not exist: {agent_name}')
        if isinstance(agent.params, dict) and agent.params.get('is_demo') is True:
            raise ValueError('Unable to delete demo object')
        agent.deleted_at = datetime.datetime.now()
        db.session.commit()

    def get_completion(
            self,
            agent: db.Agents,
            messages: List[Dict[str, str]],
            project_name: str = 'mindsdb',
            tools: List[BaseTool] = None,
            stream: bool = False) -> Union[Iterator[object], pd.DataFrame]:
        """
        Queries an agent to get a completion.

        Parameters:
            agent (db.Agents): Existing agent to get completion from
            messages (List[Dict[str, str]]): Chat history to send to the agent
            project_name (str): Project the agent belongs to (default mindsdb)
            tools (List[BaseTool]): Tools to use while getting the completion
            stream (bool): Whether to stream the response

        Returns:
            response (Union[Iterator[object], pd.DataFrame]): Completion as a DataFrame or iterator of completion chunks

        Raises:
            ValueError: Agent's model does not exist.
        """
        if stream:
            return self._get_completion_stream(
                agent,
                messages,
                project_name=project_name,
                tools=tools
            )
        from .langchain_agent import LangchainAgent

        model, provider = self.check_model_provider(agent.model_name, agent.provider)
        # update old agents
        if agent.provider is None and provider is not None:
            agent.provider = provider
            db.session.commit()

        lang_agent = LangchainAgent(agent, model)
        return lang_agent.get_completion(messages)

    def _get_completion_stream(
            self,
            agent: db.Agents,
            messages: List[Dict[str, str]],
            project_name: str = 'mindsdb',
            tools: List[BaseTool] = None) -> Iterator[object]:
        '''
        Queries an agent to get a stream of completion chunks.

        Parameters:
            agent (db.Agents): Existing agent to get completion from
            messages (List[Dict[str, str]]): Chat history to send to the agent
            trace_id (str): ID of Langfuse trace to use
            observation_id (str): ID of parent Langfuse observation to use
            project_name (str): Project the agent belongs to (default mindsdb)
            tools (List[BaseTool]): Tools to use while getting the completion

        Returns:
            chunks (Iterator[object]): Completion chunks as an iterator

        Raises:
            ValueError: Agent's model does not exist.
        '''
        # For circular dependency.
        from .langchain_agent import LangchainAgent

        model, provider = self.check_model_provider(agent.model_name, agent.provider)

        # update old agents
        if agent.provider is None and provider is not None:
            agent.provider = provider
            db.session.commit()

        lang_agent = LangchainAgent(agent, model=model)
        return lang_agent.get_completion(messages, stream=True)



import unittest

from mindsdb.integrations.handlers.solr_handler.solr_handler import SolrHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class SolrHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "username": "demo_user",
            "password": "demo_password",
            "host": "172.22.0.4",
            "port": 8983,
            "server_path": "solr",
            "collection": "gettingstarted",
            "use_ssl": False
        }
        cls.handler = SolrHandler('test_solr_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls['type'] is not RESPONSE_TYPE.ERROR

    def test_6_describe_table(self):
        described = self.handler.get_columns("gettingstarted")
        assert described['type'] is RESPONSE_TYPE.TABLE

    def test_7_select_query(self):
        query = "SELECT * FROM gettingstarted WHERE id='apple' limit 1000"
        result = self.handler.query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE

if __name__ == '__main__':
    unittest.main()


import os
import secrets
import mimetypes
import threading
import traceback
import webbrowser
from pathlib import Path
from http import HTTPStatus
from distutils.version import LooseVersion

import requests
from flask import Flask, url_for, make_response, request, send_from_directory
from flask.json import dumps
from flask_compress import Compress
from flask_restx import Api
from werkzeug.exceptions import HTTPException

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.gui import update_static
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.agents import ns_conf as agents_ns
from mindsdb.api.http.namespaces.analysis import ns_conf as analysis_ns
from mindsdb.api.http.namespaces.auth import ns_conf as auth_ns
from mindsdb.api.http.namespaces.chatbots import ns_conf as chatbots_ns
from mindsdb.api.http.namespaces.jobs import ns_conf as jobs_ns
from mindsdb.api.http.namespaces.config import ns_conf as conf_ns
from mindsdb.api.http.namespaces.databases import ns_conf as databases_ns
from mindsdb.api.http.namespaces.default import ns_conf as default_ns, check_auth
from mindsdb.api.http.namespaces.file import ns_conf as file_ns
from mindsdb.api.http.namespaces.handlers import ns_conf as handlers_ns
from mindsdb.api.http.namespaces.knowledge_bases import ns_conf as knowledge_bases_ns
from mindsdb.api.http.namespaces.models import ns_conf as models_ns
from mindsdb.api.http.namespaces.projects import ns_conf as projects_ns
from mindsdb.api.http.namespaces.skills import ns_conf as skills_ns
from mindsdb.api.http.namespaces.sql import ns_conf as sql_ns
from mindsdb.api.http.namespaces.tab import ns_conf as tab_ns
from mindsdb.api.http.namespaces.tree import ns_conf as tree_ns
from mindsdb.api.http.namespaces.views import ns_conf as views_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.api.http.namespaces.webhooks import ns_conf as webhooks_ns
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.storage import db
from mindsdb.metrics.server import init_metrics
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.json_encoder import CustomJSONProvider
from mindsdb.utilities.ps import is_pid_listen_port, wait_func_is_true
from mindsdb.utilities.telemetry import inject_telemetry_to_static
from mindsdb.utilities.sentry import sentry_sdk  # noqa: F401
from mindsdb.utilities.otel import trace  # noqa: F401
from opentelemetry.instrumentation.flask import FlaskInstrumentor  # noqa: F401
from opentelemetry.instrumentation.requests import RequestsInstrumentor  # noqa: F401

logger = log.getLogger(__name__)


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """
    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)


def custom_output_json(data, code, headers=None):
    resp = make_response(dumps(data, cls=CustomJSONProvider), code)
    resp.headers.extend(headers or {})
    return resp


def get_last_compatible_gui_version() -> LooseVersion:
    logger.debug("Getting last compatible frontend..")
    try:
        res = requests.get('https://mindsdb-web-builds.s3.amazonaws.com/compatible-config.json', timeout=5)
    except (ConnectionError, requests.exceptions.ConnectionError) as e:
        logger.error(f"Is no connection. {e}")
        return False
    except Exception as e:
        logger.error(f"Is something wrong with getting compatible-config.json: {e}")
        return False

    if res.status_code != 200:
        logger.error(
            f"Cant get compatible-config.json: returned status code = {res.status_code}"
        )
        return False

    try:
        versions = res.json()
    except Exception as e:
        logger.error(f"Cant decode compatible-config.json: {e}")
        return False

    current_mindsdb_lv = LooseVersion(mindsdb_version)

    try:
        gui_versions = {}
        max_mindsdb_lv = None
        max_gui_lv = None
        for el in versions['mindsdb']:
            if el['mindsdb_version'] is None:
                gui_lv = LooseVersion(el['gui_version'])
            else:
                mindsdb_lv = LooseVersion(el['mindsdb_version'])
                gui_lv = LooseVersion(el['gui_version'])
                if mindsdb_lv.vstring not in gui_versions or gui_lv > gui_versions[mindsdb_lv.vstring]:
                    gui_versions[mindsdb_lv.vstring] = gui_lv
                if max_mindsdb_lv is None or max_mindsdb_lv < mindsdb_lv:
                    max_mindsdb_lv = mindsdb_lv
            if max_gui_lv is None or max_gui_lv < gui_lv:
                max_gui_lv = gui_lv

        all_mindsdb_lv = [LooseVersion(x) for x in gui_versions.keys()]
        all_mindsdb_lv.sort()

        if current_mindsdb_lv.vstring in gui_versions:
            gui_version_lv = gui_versions[current_mindsdb_lv.vstring]
        elif current_mindsdb_lv > all_mindsdb_lv[-1]:
            gui_version_lv = max_gui_lv
        else:
            lower_versions = {key: value for key, value in gui_versions.items() if LooseVersion(key) < current_mindsdb_lv}
            if len(lower_versions) == 0:
                gui_version_lv = gui_versions[all_mindsdb_lv[0].vstring]
            else:
                all_lower_versions = [LooseVersion(x) for x in lower_versions.keys()]
                gui_version_lv = gui_versions[all_lower_versions[-1].vstring]
    except Exception as e:
        logger.error(f"Error in compatible-config.json structure: {e}")
        return False

    logger.debug(f"Last compatible frontend version: {gui_version_lv}.")
    return gui_version_lv


def get_current_gui_version() -> LooseVersion:
    logger.debug("Getting current frontend version..")
    config = Config()
    static_path = Path(config['paths']['static'])
    version_txt_path = static_path.joinpath('version.txt')

    current_gui_version = None
    if version_txt_path.is_file():
        with open(version_txt_path, 'rt') as f:
            current_gui_version = f.readline()

    current_gui_lv = (
        None if current_gui_version is None else LooseVersion(current_gui_version)
    )
    logger.debug(f"Current frontend version: {current_gui_lv}.")

    return current_gui_lv


def initialize_static():
    logger.debug("Initializing static..")
    config = Config()
    last_gui_version_lv = get_last_compatible_gui_version()
    current_gui_version_lv = get_current_gui_version()
    required_gui_version = config['gui'].get('version')

    if required_gui_version is not None:
        required_gui_version_lv = LooseVersion(required_gui_version)
        success = True
        if (
            current_gui_version_lv is None
            or required_gui_version_lv != current_gui_version_lv
        ):
            logger.debug("Updating gui..")
            success = update_static(required_gui_version_lv)
    else:
        if last_gui_version_lv is False:
            return False

        # ignore versions like '23.9.2.2'
        if current_gui_version_lv is not None and len(current_gui_version_lv.version) < 3:
            if current_gui_version_lv == last_gui_version_lv:
                return True
        logger.debug("Updating gui..")
        success = update_static(last_gui_version_lv)

    db.session.close()
    return success


def initialize_app(config, no_studio):
    static_root = config['paths']['static']
    logger.debug(f"Static route: {static_root}")
    gui_exists = Path(static_root).joinpath('index.html').is_file()
    logger.debug(f"Does GUI already exist.. {'YES' if gui_exists else 'NO'}")
    init_static_thread = None
    if (
        no_studio is False
        and (
            config['gui']['autoupdate'] is True
            or gui_exists is False
        )
    ):
        init_static_thread = threading.Thread(target=initialize_static)
        init_static_thread.start()

    app, api = initialize_flask(config, init_static_thread, no_studio)
    Compress(app)
    initialize_interfaces(app)

    if os.path.isabs(static_root) is False:
        static_root = os.path.join(os.getcwd(), static_root)
    static_root = Path(static_root)

    @app.route('/', defaults={'path': ''}, methods=['GET'])
    @app.route('/<path:path>', methods=['GET'])
    def root_index(path):
        if path.startswith('api/'):
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Not found',
                'The endpoint you are trying to access does not exist on the server.'
            )

        # Normalize the path.
        full_path = os.path.normpath(os.path.join(static_root, path))

        # Check for directory traversal attacks.
        if not full_path.startswith(str(static_root)):
            return http_error(
                HTTPStatus.FORBIDDEN,
                'Forbidden',
                'You are not allowed to access the requested resource.'
            )

        if os.path.isfile(full_path):
            return send_from_directory(static_root, path)
        else:
            return send_from_directory(static_root, 'index.html')

    protected_namespaces = [
        tab_ns,
        utils_ns,
        conf_ns,
        file_ns,
        sql_ns,
        analysis_ns,
        handlers_ns,
        tree_ns,
        projects_ns,
        databases_ns,
        views_ns,
        models_ns,
        chatbots_ns,
        skills_ns,
        agents_ns,
        jobs_ns,
        knowledge_bases_ns
    ]

    for ns in protected_namespaces:
        api.add_namespace(ns)
    api.add_namespace(default_ns)
    api.add_namespace(auth_ns)
    api.add_namespace(webhooks_ns)

    @api.errorhandler(Exception)
    def handle_exception(e):
        logger.error(f"http exception: {e}")
        # pass through HTTP errors
        # NOTE flask_restx require 'message', also it modyfies 'application/problem+json' to 'application/json'
        if isinstance(e, HTTPException):
            return ({
                'title': e.name,
                'detail': e.description,
                'message': e.description
            }, e.code, {
                'Content-Type': 'application/problem+json'
            })
        return ({
            'title': getattr(type(e), "__name__") or "Unknown error",
            'detail': str(e),
            'message': str(e)
        }, 500, {
            'Content-Type': 'application/problem+json'
        })

    @app.teardown_appcontext
    def remove_session(*args, **kwargs):
        db.session.remove()

    @app.before_request
    def before_request():
        logger.debug(f"HTTP {request.method}: {request.path}")
        ctx.set_default()
        config = Config()

        # region routes where auth is required
        if (
            config['auth']['http_auth_enabled'] is True
            and any(request.path.startswith(f'/api{ns.path}') for ns in protected_namespaces)
            and check_auth() is False
        ):
            return http_error(
                HTTPStatus.UNAUTHORIZED, 'Unauthorized',
                'Authorization is required to complete the request'
            )
        # endregion

        company_id = request.headers.get('company-id')
        user_class = request.headers.get('user-class')

        try:
            email_confirmed = int(request.headers.get('email-confirmed', 1))
        except Exception:
            email_confirmed = 1

        try:
            user_id = int(request.headers.get('user-id', 0))
        except Exception:
            user_id = 0

        try:
            session_id = request.cookies.get('session')
        except Exception:
            session_id = "unknown"

        if company_id is not None:
            try:
                company_id = int(company_id)
            except Exception as e:
                logger.error(
                    f"Cloud not parse company id: {company_id} | exception: {e}"
                )
                company_id = None

        if user_class is not None:
            try:
                user_class = int(user_class)
            except Exception as e:
                logger.error(
                    f"Cloud not parse user_class: {user_class} | exception: {e}"
                )
                user_class = 0
        else:
            user_class = 0

        ctx.user_id = user_id
        ctx.session_id = session_id
        ctx.company_id = company_id
        ctx.user_class = user_class
        ctx.email_confirmed = email_confirmed

    # Wait for static initialization.
    if not no_studio and init_static_thread is not None:
        init_static_thread.join()

    logger.debug("Done initializing app.")
    return app


def initialize_flask(config, init_static_thread, no_studio):
    logger.debug("Initializing flask..")
    # region required for windows https://github.com/mindsdb/mindsdb/issues/2526
    mimetypes.add_type('text/css', '.css')
    mimetypes.add_type('text/javascript', '.js')
    # endregion

    kwargs = {}
    if no_studio is not True:
        static_path = os.path.join(config['paths']['static'], 'static/')
        if os.path.isabs(static_path) is False:
            static_path = os.path.join(os.getcwd(), static_path)
        kwargs["static_url_path"] = "/static"
        kwargs["static_folder"] = static_path
        logger.debug(f"Static path: {static_path}")

    app = Flask(__name__, **kwargs)
    init_metrics(app)

    # Instrument Flask app for OpenTelemetry
    FlaskInstrumentor().instrument_app(app)
    RequestsInstrumentor().instrument()

    app.config['SECRET_KEY'] = os.environ.get('FLASK_SECRET_KEY', secrets.token_hex(32))
    app.config['SESSION_COOKIE_NAME'] = 'session'
    app.config['PERMANENT_SESSION_LIFETIME'] = config['auth']['http_permanent_session_lifetime']
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 60
    app.config['SWAGGER_HOST'] = 'http://localhost:8000/mindsdb'
    app.json = CustomJSONProvider()

    authorizations = {
        'apikey': {
            'type': 'session',
            'in': 'query',
            'name': 'session'
        }
    }

    logger.debug("Creating swagger API..")
    api = Swagger_Api(
        app,
        authorizations=authorizations,
        security=['apikey'],
        url_prefix=':8000',
        prefix='/api',
        doc='/doc/'
    )

    api.representations['application/json'] = custom_output_json

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    # NOTE rewrite it, that hotfix to see GUI link
    if not no_studio:
        if host in ("", "0.0.0.0"):
            url = f"http://127.0.0.1:{port}/"
        else:
            url = f"http://{host}:{port}/"
        logger.info(f" - GUI available at {url}")

        pid = os.getpid()
        x = threading.Thread(target=_open_webbrowser, args=(url, pid, port, init_static_thread, config['paths']['static']), daemon=True)
        x.start()

    return app, api


def initialize_interfaces(app):
    app.integration_controller = integration_controller
    app.database_controller = DatabaseController()
    app.file_controller = FileController()
    config = Config()
    app.config_obj = config


def _open_webbrowser(url: str, pid: int, port: int, init_static_thread, static_folder):
    """Open webbrowser with url when http service is started.

    If some error then do nothing.
    """
    if init_static_thread is not None:
        init_static_thread.join()
    inject_telemetry_to_static(static_folder)
    try:
        is_http_active = wait_func_is_true(
            func=is_pid_listen_port, timeout=15, pid=pid, port=port
        )
        if is_http_active:
            webbrowser.open(url)
    except Exception as e:
        logger.error(f"Failed to open {url} in webbrowser with exception {e}")
        logger.error(traceback.format_exc())
    db.session.close()



from typing import List
from contextlib import contextmanager

import boto3
import duckdb
from duckdb import HTTPException
from mindsdb_sql_parser import parse_sql
import pandas as pd
from typing import Text, Dict, Optional
from botocore.exceptions import ClientError

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert, Star, Constant

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

from mindsdb.integrations.libs.api_handler import APIResource, APIHandler
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator

logger = log.getLogger(__name__)


class ListFilesTable(APIResource):

    def list(self,
             targets: List[str] = None,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             *args, **kwargs) -> pd.DataFrame:

        buckets = None
        for condition in conditions:
            if condition.column == 'bucket':
                if condition.op == FilterOperator.IN:
                    buckets = condition.value
                elif condition.op == FilterOperator.EQUAL:
                    buckets = [condition.value]
                condition.applied = True

        data = []
        for obj in self.handler.get_objects(limit=limit, buckets=buckets):
            path = obj['Key']
            path = path.replace('`', '')
            item = {
                'path': path,
                'bucket': obj['Bucket'],
                'name': path[path.rfind('/') + 1:],
                'extension': path[path.rfind('.') + 1:]
            }

            data.append(item)

        return pd.DataFrame(data=data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        return ["path", "name", "extension", "bucket", "content"]


class FileTable(APIResource):

    def list(self, targets: List[str] = None, table_name=None, *args, **kwargs) -> pd.DataFrame:
        return self.handler.read_as_table(table_name)

    def add(self, data, table_name=None):
        df = pd.DataFrame(data)
        return self.handler.add_data_to_table(table_name, df)


class S3Handler(APIHandler):
    """
    This handler handles connection and execution of the SQL statements on AWS S3.
    """

    name = 's3'
    # TODO: Can other file formats be supported?
    supported_file_formats = ['csv', 'tsv', 'json', 'parquet']

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs):
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the AWS (S3) account.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True
        self.bucket = self.connection_data.get('bucket')
        self._regions = {}

        self._files_table = ListFilesTable(self)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Establishes a connection to the AWS (S3) account.

        Raises:
            ValueError: If the required connection parameters are not provided.

        Returns:
            boto3.client: A client object to the AWS (S3) account.
        """
        if self.is_connected is True:
            return self.connection

        # Validate mandatory parameters.
        if not all(key in self.connection_data for key in ['aws_access_key_id', 'aws_secret_access_key']):
            raise ValueError('Required parameters (aws_access_key_id, aws_secret_access_key) must be provided.')

        # Connect to S3 and configure mandatory credentials.
        self.connection = self._connect_boto3()
        self.is_connected = True

        return self.connection

    @contextmanager
    def _connect_duckdb(self, bucket):
        """
        Creates temporal duckdb database which is able to connect to the AWS (S3) account.
        Have to be used as context manager

        Returns:
            DuckDBPyConnection
        """
        # Connect to S3 via DuckDB.
        duckdb_conn = duckdb.connect(":memory:")
        try:
            duckdb_conn.execute("INSTALL httpfs")
        except HTTPException as http_error:
            logger.debug(f"Error installing the httpfs extension, {http_error}! Forcing installation.")
            duckdb_conn.execute("FORCE INSTALL httpfs")

        duckdb_conn.execute("LOAD httpfs")

        # Configure mandatory credentials.
        duckdb_conn.execute(f"SET s3_access_key_id='{self.connection_data['aws_access_key_id']}'")
        duckdb_conn.execute(f"SET s3_secret_access_key='{self.connection_data['aws_secret_access_key']}'")

        # Configure optional parameters.
        if 'aws_session_token' in self.connection_data:
            duckdb_conn.execute(f"SET s3_session_token='{self.connection_data['aws_session_token']}'")

        # detect region for bucket
        if bucket not in self._regions:
            client = self.connect()
            self._regions[bucket] = client.get_bucket_location(Bucket=bucket)['LocationConstraint']

        region = self._regions[bucket]
        duckdb_conn.execute(f"SET s3_region='{region}'")

        try:
            yield duckdb_conn
        finally:
            duckdb_conn.close()

    def _connect_boto3(self) -> boto3.client:
        """
        Establishes a connection to the AWS (S3) account.

        Returns:
            boto3.client: A client object to the AWS (S3) account.
        """
        # Configure mandatory credentials.
        config = {
            'aws_access_key_id': self.connection_data['aws_access_key_id'],
            'aws_secret_access_key': self.connection_data['aws_secret_access_key']
        }

        # Configure optional parameters.
        if 'aws_session_token' in self.connection_data:
            config['aws_session_token'] = self.connection_data['aws_session_token']

        client = boto3.client('s3', **config)

        # check connection
        if self.bucket is not None:
            client.head_bucket(Bucket=self.bucket)
        else:
            client.list_buckets()

        return client

    def disconnect(self):
        """
        Closes the connection to the AWS (S3) account if it's currently open.
        """
        if not self.is_connected:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the S3 bucket.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        # Check connection via boto3.
        try:
            self._connect_boto3()
            response.success = True
        except (ClientError, ValueError) as e:
            logger.error(f'Error connecting to S3 with the given credentials, {e}!')
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def _get_bucket(self, key):
        if self.bucket is not None:
            return self.bucket, key

        # get bucket from first part of the key
        ar = key.split('/')
        return ar[0], '/'.join(ar[1:])

    def read_as_table(self, key) -> pd.DataFrame:
        """
        Read object as dataframe. Uses duckdb
        """
        bucket, key = self._get_bucket(key)

        with self._connect_duckdb(bucket) as connection:

            cursor = connection.execute(f"SELECT * FROM 's3://{bucket}/{key}'")

            return cursor.fetchdf()

    def _read_as_content(self, key) -> None:
        """
        Read object as content
        """
        bucket, key = self._get_bucket(key)

        client = self.connect()

        obj = client.get_object(Bucket=bucket, Key=key)
        content = obj['Body'].read()
        return content

    def add_data_to_table(self, key, df) -> None:
        """
        Writes the table to a file in the S3 bucket.

        Raises:
            CatalogException: If the table does not exist in the DuckDB connection.
        """

        # Check if the file exists in the S3 bucket.
        bucket, key = self._get_bucket(key)

        try:
            client = self.connect()
            client.head_object(Bucket=bucket, Key=key)
        except ClientError as e:
            logger.error(f'Error querying the file {key} in the bucket {bucket}, {e}!')
            raise e

        with self._connect_duckdb(bucket) as connection:
            # copy
            connection.execute(f"CREATE TABLE tmp_table AS SELECT * FROM 's3://{bucket}/{key}'")

            # insert
            connection.execute("INSERT INTO tmp_table BY NAME SELECT * FROM df")

            # upload
            connection.execute(f"COPY tmp_table TO 's3://{bucket}/{key}'")

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Raises:
            ValueError: If the file format is not supported or the file does not exist in the S3 bucket.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        self.connect()

        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]

            if table_name == 'files':
                table = self._files_table
                df = table.select(query)

                # add content
                has_content = False
                for target in query.targets:
                    if isinstance(target, Identifier) and target.parts[-1].lower() == 'content':
                        has_content = True
                        break
                if has_content:
                    df['content'] = df['path'].apply(self._read_as_content)
            else:
                extension = table_name.split('.')[-1]
                if extension not in self.supported_file_formats:
                    logger.error(f'The file format {extension} is not supported!')
                    raise ValueError(f'The file format {extension} is not supported!')

                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_frame=df
            )
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            response = Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        query_ast = parse_sql(query)
        return self.query(query_ast)

    def get_objects(self, limit=None, buckets=None) -> List[dict]:
        client = self.connect()
        if self.bucket is not None:
            add_bucket_to_name = False
            scan_buckets = [self.bucket]
        else:
            add_bucket_to_name = True
            scan_buckets = [b['Name'] for b in client.list_buckets()['Buckets']]

        objects = []
        for bucket in scan_buckets:
            if buckets is not None and bucket not in buckets:
                continue

            resp = client.list_objects_v2(Bucket=bucket)
            if 'Contents' not in resp:
                continue

            for obj in resp['Contents']:
                if obj.get('StorageClass', 'STANDARD') != 'STANDARD':
                    continue

                obj['Bucket'] = bucket
                if add_bucket_to_name:
                    # bucket is part of the name
                    obj['Key'] = f'{bucket}/{obj["Key"]}'
                objects.append(obj)
            if limit is not None and len(objects) >= limit:
                break

        return objects

    def get_tables(self) -> Response:
        """
        Retrieves a list of tables (objects) in the S3 bucket.

        Each object is considered a table. Only the supported file formats are considered as tables.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """

        # Get only the supported file formats.
        # Wrap the object names with backticks to prevent SQL syntax errors.
        supported_names = [
            f"`{obj['Key']}`"
            for obj in self.get_objects()
            if obj['Key'].split('.')[-1] in self.supported_file_formats
        ]

        # virtual table with list of files
        supported_names.insert(0, 'files')

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                supported_names,
                columns=['table_name']
            )
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Retrieves column details for a specified table (object) in the S3 bucket.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        """
        query = Select(
            targets=[Star()],
            from_table=Identifier(parts=[table_name]),
            limit=Constant(1)
        )

        result = self.query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': result.data_frame.columns,
                    'data_type': [data_type if data_type != 'object' else 'string' for data_type in result.data_frame.dtypes]
                }
            )
        )

        return response



import datetime

from alembic.autogenerate import produce_migrations, render, api
from alembic import context
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Boolean, Index

# required for code execution
from alembic import op  # noqa
import sqlalchemy as sa  # noqa

import mindsdb.interfaces.storage.db    # noqa
from mindsdb.interfaces.storage.db import Json, Array
from mindsdb.utilities import log

logger = log.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '17c3d2384711'
down_revision = None
branch_labels = None
depends_on = None

# ========================================== current database state ========================================


class Base:
    __allow_unmapped__ = True


Base = declarative_base(cls=Base)

# Source: https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable


class Semaphor(Base):
    __tablename__ = 'semaphor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    entity_type = Column('entity_type', String)
    entity_id = Column('entity_id', Integer)
    action = Column(String)
    company_id = Column(Integer)
    uniq_const = UniqueConstraint('entity_type', 'entity_id')


class Datasource(Base):
    __tablename__ = 'datasource'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String)
    data = Column(String)  # Including, e.g. the query used to create it and even the connection info when there's no integration associated with it -- A JSON
    creation_info = Column(String)
    analysis = Column(String)  # A JSON
    company_id = Column(Integer)
    mindsdb_version = Column(String)
    datasources_version = Column(String)
    integration_id = Column(Integer)


class Predictor(Base):
    __tablename__ = 'predictor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String)
    data = Column(Json)  # A JSON -- should be everything returned by `get_model_data`, I think
    to_predict = Column(Array)
    company_id = Column(Integer)
    mindsdb_version = Column(String)
    native_version = Column(String)
    datasource_id = Column(Integer)
    is_custom = Column(Boolean)     # to del
    learn_args = Column(Json)
    update_status = Column(String, default='up_to_date')

    json_ai = Column(Json, nullable=True)
    code = Column(String, nullable=True)
    lightwood_version = Column(String, nullable=True)
    dtype_dict = Column(Json, nullable=True)


class AITable(Base):
    __tablename__ = 'ai_table'
    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String)
    integration_name = Column(String)
    integration_query = Column(String)
    query_fields = Column(Json)
    predictor_name = Column(String)
    predictor_columns = Column(Json)
    company_id = Column(Integer)


class Log(Base):
    __tablename__ = 'log'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    log_type = Column(String)  # log, info, warning, traceback etc
    source = Column(String)  # file + line
    company_id = Column(Integer)
    payload = Column(String)
    created_at_index = Index("some_index", "created_at_index")


class Integration(Base):
    __tablename__ = 'integration'
    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String, nullable=False)
    data = Column(Json)
    company_id = Column(Integer)


class Stream(Base):
    __tablename__ = 'stream'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    stream_in = Column(String, nullable=False)
    stream_out = Column(String, nullable=False)
    anomaly_stream = Column(String)
    integration = Column(String)
    predictor = Column(String, nullable=False)
    company_id = Column(Integer)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    type = Column(String, default='unknown')
    connection_info = Column(Json, default={})
    learning_params = Column(Json, default={})
    learning_threshold = Column(Integer, default=0)


# ====================================================================================================


def upgrade():
    '''
       First migration.
       Generates a migration script by difference between model and database and executes it
    '''

    target_metadata = Base.metadata

    mc = context.get_context()

    migration_script = produce_migrations(mc, target_metadata)

    autogen_context = api.AutogenContext(
        mc, autogenerate=True
    )

    # Seems to be the only way to apply changes to the database
    template_args = {}
    render._render_python_into_templatevars(
        autogen_context, migration_script, template_args
    )

    code = template_args['upgrades']
    code = code.replace('\n    ', '\n')
    logger.info('\nPerforming database changes:')
    logger.info(code)
    exec(code)


def downgrade():

    # We don't know state to downgrade
    raise NotImplementedError()



from collections import OrderedDict
from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.handlers.lightdash_handler.api import Lightdash
from mindsdb.integrations.handlers.lightdash_handler.lightdash_tables import (
    UserTable,
    UserAbilityTable,
    OrgTable,
    OrgProjectsTable,
    OrgMembersTable,
    ProjectTable,
    WarehouseConnectionTable,
    DBTConnectionTable,
    DBTEnvironmentVarsTable,
    ChartsTable,
    SpacesTable,
    AccessTable,
    ValidationTable,
    DashboardsTable,
    QueriesTable,
    ChartHistoryTable,
    ChartConfigTable,
    ChartAdditionalMetricsTable,
    ChartTableCalculationsTable,
    SchedulerLogsTable,
    SchedulerTable,
    SchedulerJobsTable,
    SchedulerJobStatus,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse


class LightdashHandler(APIHandler):

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(name)
        self.connection = None
        self.is_connected = False
        self.api_key = kwargs.get("connection_data", {}).get("api_key", "")
        self.base_url = kwargs.get("connection_data", {}).get("base_url", "")
        _tables = [
            UserTable,
            UserAbilityTable,
            OrgTable,
            OrgProjectsTable,
            OrgMembersTable,
            ProjectTable,
            WarehouseConnectionTable,
            DBTConnectionTable,
            DBTEnvironmentVarsTable,
            ChartsTable,
            SpacesTable,
            AccessTable,
            ValidationTable,
            DashboardsTable,
            QueriesTable,
            ChartHistoryTable,
            ChartConfigTable,
            ChartAdditionalMetricsTable,
            ChartTableCalculationsTable,
            SchedulerLogsTable,
            SchedulerTable,
            SchedulerJobsTable,
            SchedulerJobStatus,
        ]
        for Table in _tables:
            self._register_table(Table.name, Table(self))
        self.connect()

    def connect(self) -> Lightdash:
        self.connection = Lightdash(self.base_url, self.api_key)
        return self.connection

    def check_connection(self) -> StatusResponse:
        resp = StatusResponse(False)
        if self.connection and not self.connection.is_connected():
            resp.error = "Client not connected"
        else:
            resp.success = True
        return resp

    def native_query(self, query: str) -> StatusResponse:
        ast = parse_sql(query)
        return self.query(ast)


connection_args = OrderedDict(
    api_key={
        'type': ARG_TYPE.STR,
        'description': 'API Token for accessing Lightdash instance',
        'required': True,
        'label': 'API Key',
    },
    base_url={
        'type': ARG_TYPE.STR,
        'description': 'Base URL of Lightdash instance',
        'required': True,
        'label': 'Base URL',
    }
)


connection_args_example = OrderedDict(
    api_key='23d6b9e0c2fab7eba2e8b7e452cead47',
    base_url='http://localhost:8080/'
)



from mindsdb.api.executor.controllers import SessionController
from mindsdb.api.executor.command_executor import ExecuteCommands


def run_sql_command(request_env, ast_query):
    sql_session = SessionController()
    sql_session.database = request_env.get('database', 'mindsdb')

    command_executor = ExecuteCommands(sql_session)
    ret = command_executor.execute_command(ast_query)
    if ret.error_code is not None:
        raise Exception(ret.error_message)

    if ret.data is None:
        # return no data
        return []

    return list(ret.data.get_records())



import copy
import traceback
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Data, Identifier
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.interfaces.storage import db

from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.command_executor import ExecuteCommands

from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.utilities import log
from mindsdb.interfaces.tasks.task import BaseTask
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


class TriggerTask(BaseTask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.command_executor = None
        self.query = None

        # callback might be without context
        self._ctx_dump = ctx.dump()

    def run(self, stop_event):
        trigger = db.Triggers.query.get(self.object_id)

        # parse query
        self.query = parse_sql(trigger.query_str)

        session = SessionController()

        # prepare executor
        project_controller = ProjectController()
        project = project_controller.get(trigger.project_id)

        session.database = project.name

        self.command_executor = ExecuteCommands(session)

        # subscribe
        database = session.integration_controller.get_by_id(trigger.database_id)
        data_handler = session.integration_controller.get_data_handler(database['name'])

        columns = trigger.columns
        if columns is not None:
            if columns == '':
                columns = None
            else:
                columns = columns.split('|')

        data_handler.subscribe(stop_event, self._callback, trigger.table_name, columns=columns)

    def _callback(self, row, key=None):
        logger.debug(f'trigger call: {row}, {key}')

        # set up environment
        ctx.load(self._ctx_dump)

        try:
            if key is not None:
                row.update(key)
            table = [
                row
            ]

            # inject data to query
            query = copy.deepcopy(self.query)

            def find_table(node, is_table, **kwargs):

                if is_table:
                    if (
                            isinstance(node, Identifier)
                            and len(node.parts) == 1
                            and node.parts[0] == 'TABLE_DELTA'
                    ):
                        # replace with data
                        return Data(table, alias=node.alias)

            query_traversal(query, find_table)

            # exec query
            ret = self.command_executor.execute_command(query)
            if ret.error_code is not None:
                self.set_error(ret.error_message)

        except Exception:
            self.set_error(str(traceback.format_exc()))

        db.session.commit()
